import re
import sys
import types
from collections import defaultdict
from contextlib import contextmanager

from .._compat import (PY2, basestring, hashlib_md5, integer_types, iteritems,
                       iterkeys, with_metaclass)
from .._globals import IDENTITY
from ..connection import ConnectionPool
from ..exceptions import NotOnNOSQLError
from ..helpers.classes import (SQLALL, ExecutionHandler, NullDriver, Reference,
                               SQLCustomType)
from ..helpers.methods import merge_tablemaps, use_common_filters, xorify
from ..helpers.regex import REGEX_SELECT_AS_PARSER, REGEX_TABLE_DOT_FIELD
from ..migrator import Migrator
from ..objects import (Expression, Field, IterRows, LazyReferenceGetter,
                       LazySet, Query, Rows, Select, Table, VirtualCommand)
from ..utils import deprecated
from . import AdapterMeta, with_connection, with_connection_or_raise

CALLABLETYPES = (
    types.LambdaType,
    types.FunctionType,
    types.BuiltinFunctionType,
    types.MethodType,
    types.BuiltinMethodType,
)


class BaseAdapter(with_metaclass(AdapterMeta, ConnectionPool)):
    dbengine = "None"
    drivers = ()
    uploads_in_blob = False
    support_distributed_transaction = False

    def __init__(
        self,
        db,
        uri,
        pool_size=0,
        folder=None,
        db_codec="UTF-8",
        credential_decoder=IDENTITY,
        driver_args={},
        adapter_args={},
        after_connection=None,
        entity_quoting=False,
    ):
        super(BaseAdapter, self).__init__()
        self._load_dependencies()
        self.db = db
        self.uri = uri
        self.pool_size = pool_size
        self.folder = folder
        self.db_codec = db_codec
        self.credential_decoder = credential_decoder
        self.driver_args = driver_args
        self.adapter_args = adapter_args
        self.expand = self._expand
        self._after_connection = after_connection
        self.set_connection(None)
        self.find_driver()
        self._initialize_()

    def _load_dependencies(self):
        from ..dialects import dialects
        from ..parsers import parsers
        from ..representers import representers

        self.dialect = dialects.get_for(self)
        self.parser = parsers.get_for(self)
        self.representer = representers.get_for(self)

    def _initialize_(self):
        self._find_work_folder()

    @property
    def types(self):
        return self.dialect.types

    @property
    def _available_drivers(self):
        return [
            driver
            for driver in self.drivers
            if driver in iterkeys(self.db._drivers_available)
        ]

    def _driver_from_uri(self):
        rv = None
        if self.uri:
            items = self.uri.split("://", 1)[0].split(":")
            rv = items[1] if len(items) > 1 else None
        return rv

    def find_driver(self):
        if getattr(self, "driver", None) is not None:
            return
        requested_driver = self._driver_from_uri() or self.adapter_args.get("driver")
        if requested_driver:
            if requested_driver in self._available_drivers:
                self.driver_name = requested_driver
                self.driver = self.db._drivers_available[requested_driver]
            else:
                raise RuntimeError("Driver %s is not available" % requested_driver)
        elif self._available_drivers:
            self.driver_name = self._available_drivers[0]
            self.driver = self.db._drivers_available[self.driver_name]
        else:
            raise RuntimeError(
                "No driver of supported ones %s is available" % str(self.drivers)
            )

    def connector(self):
        return self.driver.connect(self.driver_args)

    def test_connection(self):
        pass

    @with_connection
    def close_connection(self):
        rv = self.connection.close()
        self.set_connection(None)
        return rv

    def tables(self, *queries):
        tables = dict()
        for query in queries:
            if isinstance(query, Field):
                key = query.tablename
                if tables.get(key, query.table) is not query.table:
                    raise ValueError("Name conflict in table list: %s" % key)
                tables[key] = query.table
            elif isinstance(query, (Expression, Query)):
                tmp = [x for x in (query.first, query.second) if x is not None]
                tables = merge_tablemaps(tables, self.tables(*tmp))
        return tables

    def get_table(self, *queries):
        tablemap = self.tables(*queries)
        if len(tablemap) == 1:
            return tablemap.popitem()[1]
        elif len(tablemap) < 1:
            raise RuntimeError("No table selected")
        else:
            raise RuntimeError("Too many tables selected (%s)" % str(list(tablemap)))

    def common_filter(self, query, tablist):
        tenant_fieldname = self.db._request_tenant
        for table in tablist:
            if isinstance(table, basestring):
                table = self.db[table]
            # deal with user provided filters
            if table._common_filter is not None:
                query = query & table._common_filter(query)
            # deal with multi_tenant filters
            if tenant_fieldname in table:
                default = table[tenant_fieldname].default
                if default is not None:
                    newquery = table[tenant_fieldname] == default
                    if query is None:
                        query = newquery
                    else:
                        query = query & newquery
        return query

    def _expand(self, expression, field_type=None, colnames=False, query_env={}):
        return str(expression)

    def expand_all(self, fields, tabledict):
        new_fields = []
        append = new_fields.append
        for item in fields:
            if isinstance(item, SQLALL):
                new_fields += item._table
            elif isinstance(item, str):
                m = REGEX_TABLE_DOT_FIELD.match(item)
                if m:
                    tablename, fieldname = m.groups()
                    append(self.db[tablename][fieldname])
                else:
                    append(Expression(self.db, lambda item=item: item))
            else:
                append(item)
        # ## if no fields specified take them all from the requested tables
        if not new_fields:
            for table in tabledict.values():
                for field in table:
                    append(field)
        return new_fields

    def parse_value(self, value, field_itype, field_type, blob_decode=True):
        # [Note - gi0baro] I think next if block can be (should be?) avoided
        if field_type != "blob" and isinstance(value, str):
            try:
                value = value.decode(self.db._db_codec)
            except Exception:
                pass
        if PY2 and isinstance(value, unicode):
            value = value.encode("utf-8")
        if isinstance(field_type, SQLCustomType):
            value = field_type.decoder(value)
        if not isinstance(field_type, str) or value is None:
            return value
        elif field_type == "blob" and not blob_decode:
            return value
        else:
            return self.parser.parse(value, field_itype, field_type)

    def _add_operators_to_parsed_row(self, rid, table, row):
        for key, record_operator in iteritems(self.db.record_operators):
            setattr(row, key, record_operator(row, table, rid))
        if table._db._lazy_tables:
            row["__get_lazy_reference__"] = LazyReferenceGetter(table, rid)

    def _add_reference_sets_to_parsed_row(self, rid, table, tablename, row):
        for rfield in table._referenced_by:
            referee_link = self.db._referee_name and self.db._referee_name % dict(
                table=rfield.tablename, field=rfield.name
            )
            if referee_link and referee_link not in row and referee_link != tablename:
                row[referee_link] = LazySet(rfield, rid)

    def _regex_select_as_parser(self, colname):
        return re.search(REGEX_SELECT_AS_PARSER, colname)

    def _parse(
        self,
        row,
        tmps,
        fields,
        colnames,
        blob_decode,
        cacheable,
        fields_virtual,
        fields_lazy,
    ):
        new_row = defaultdict(self.db.Row)
        extras = self.db.Row()
        #: let's loop over columns
        for (j, colname) in enumerate(colnames):
            value = row[j]
            tmp = tmps[j]
            tablename = None
            #: do we have a real column?
            if tmp:
                (tablename, fieldname, table, field, ft, fit) = tmp
                colset = new_row[tablename]
                #: parse value
                value = self.parse_value(value, fit, ft, blob_decode)
                if field.filter_out:
                    value = field.filter_out(value)
                colset[fieldname] = value
                #! backward compatibility
                if ft == "id" and fieldname != "id" and "id" not in table.fields:
                    colset["id"] = value
                #: additional parsing for 'id' fields
                if ft == "id" and not cacheable:
                    self._add_operators_to_parsed_row(value, table, colset)
                    #: table may be 'nested_select' which doesn't have '_referenced_by'
                    if hasattr(table, "_referenced_by"):
                        self._add_reference_sets_to_parsed_row(
                            value, table, tablename, colset
                        )
            #: otherwise we set the value in extras
            else:
                #: fields[j] may be None if only 'colnames' was specified in db.executesql()
                f_itype, ftype = (
                    fields[j] and [fields[j]._itype, fields[j].type] or [None, None]
                )
                value = self.parse_value(value, f_itype, ftype, blob_decode)
                extras[colname] = value
                if not fields[j]:
                    new_row[colname] = value
                else:
                    new_column_match = self._regex_select_as_parser(colname)
                    if new_column_match is not None:
                        new_column_name = new_column_match.group(1)
                        new_row[new_column_name] = value
        #: add extras if needed (eg. operations results)
        if extras:
            new_row["_extra"] = extras
        #: add virtuals
        new_row = self.db.Row(**new_row)
        for tablename in fields_virtual.keys():
            for f, v in fields_virtual[tablename][1]:
                try:
                    new_row[tablename][f] = v.f(new_row)
                except (AttributeError, KeyError):
                    pass  # not enough fields to define virtual field
            for f, v in fields_lazy[tablename][1]:
                try:
                    new_row[tablename][f] = v.handler(v.f, new_row)
                except (AttributeError, KeyError):
                    pass  # not enough fields to define virtual field
        return new_row

    def _parse_expand_colnames(self, fieldlist):
        """
        - Expand a list of colnames into a list of
          (tablename, fieldname, table_obj, field_obj, field_type)
        - Create a list of table for virtual/lazy fields
        """
        fields_virtual = {}
        fields_lazy = {}
        tmps = []
        for field in fieldlist:
            if not isinstance(field, Field):
                tmps.append(None)
                continue
            table = field.table
            tablename, fieldname = table._tablename, field.name
            ft = field.type
            fit = field._itype
            tmps.append((tablename, fieldname, table, field, ft, fit))
            if tablename not in fields_virtual:
                fields_virtual[tablename] = (
                    table,
                    [(f.name, f) for f in table._virtual_fields],
                )
                fields_lazy[tablename] = (
                    table,
                    [(f.name, f) for f in table._virtual_methods],
                )
        return (fields_virtual, fields_lazy, tmps)

    def parse(self, rows, fields, colnames, blob_decode=True, cacheable=False):
        (fields_virtual, fields_lazy, tmps) = self._parse_expand_colnames(fields)
        new_rows = [
            self._parse(
                row,
                tmps,
                fields,
                colnames,
                blob_decode,
                cacheable,
                fields_virtual,
                fields_lazy,
            )
            for row in rows
        ]
        rowsobj = self.db.Rows(self.db, new_rows, colnames, rawrows=rows, fields=fields)
        # Old style virtual fields
        for tablename, tmp in fields_virtual.items():
            table = tmp[0]
            # ## old style virtual fields
            for item in table.virtualfields:
                try:
                    rowsobj = rowsobj.setvirtualfields(**{tablename: item})
                except (KeyError, AttributeError):
                    # to avoid breaking virtualfields when partial select
                    pass
        return rowsobj

    def iterparse(self, sql, fields, colnames, blob_decode=True, cacheable=False):
        """
        Iterator to parse one row at a time.
        It doesn't support the old style virtual fields
        """
        return IterRows(self.db, sql, fields, colnames, blob_decode, cacheable)

    def adapt(self, value):
        return value

    def represent(self, obj, field_type):
        if isinstance(obj, CALLABLETYPES):
            obj = obj()
        return self.representer.represent(obj, field_type)

    def _drop_table_cleanup(self, table):
        del self.db[table._tablename]
        del self.db.tables[self.db.tables.index(table._tablename)]
        self.db._remove_references_to(table)

    def drop_table(self, table, mode=""):
        self._drop_table_cleanup(table)

    def rowslice(self, rows, minimum=0, maximum=None):
        return rows

    def sqlsafe_table(self, tablename, original_tablename=None):
        return tablename

    def sqlsafe_field(self, fieldname):
        return fieldname


class DebugHandler(ExecutionHandler):
    def before_execute(self, command):
        self.adapter.db.logger.debug("SQL: %s" % command)


class SQLAdapter(BaseAdapter):
    commit_on_alter_table = False
    # [Note - gi0baro] can_select_for_update should be deprecated and removed
    can_select_for_update = True
    execution_handlers = []
    migrator_cls = Migrator

    def __init__(self, *args, **kwargs):
        super(SQLAdapter, self).__init__(*args, **kwargs)
        migrator_cls = self.adapter_args.get("migrator", self.migrator_cls)
        self.migrator = migrator_cls(self)
        self.execution_handlers = list(self.db.execution_handlers)
        if self.db._debug:
            self.execution_handlers.insert(0, DebugHandler)

    def test_connection(self):
        self.execute("SELECT 1;")

    def represent(self, obj, field_type):
        if isinstance(obj, (Expression, Field)):
            return str(obj)
        return super(SQLAdapter, self).represent(obj, field_type)

    def adapt(self, obj):
        return "'%s'" % obj.replace("'", "''")

    def smart_adapt(self, obj):
        if isinstance(obj, (int, float)):
            return str(obj)
        return self.adapt(str(obj))

    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        return self.cursor.fetchone()

    def _build_handlers_for_execution(self):
        rv = []
        for handler_class in self.execution_handlers:
            rv.append(handler_class(self))
        return rv

    def filter_sql_command(self, command):
        return command

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        command = self.filter_sql_command(args[0])
        handlers = self._build_handlers_for_execution()
        for handler in handlers:
            handler.before_execute(command)
        rv = self.cursor.execute(command, *args[1:], **kwargs)
        for handler in handlers:
            handler.after_execute(command)
        return rv

    def _expand(self, expression, field_type=None, colnames=False, query_env={}):
        if isinstance(expression, Field):
            if not colnames:
                rv = expression.sqlsafe
            else:
                rv = expression.longname
            if field_type == "string" and expression.type not in (
                "string",
                "text",
                "json",
                "jsonb",
                "password",
            ):
                rv = self.dialect.cast(rv, self.types["text"], query_env)
        elif isinstance(expression, (Expression, Query)):
            first = expression.first
            second = expression.second
            op = expression.op
            optional_args = expression.optional_args or {}
            optional_args["query_env"] = query_env
            if second is not None:
                rv = op(first, second, **optional_args)
            elif first is not None:
                rv = op(first, **optional_args)
            elif isinstance(op, str):
                if op.endswith(";"):
                    op = op[:-1]
                rv = "(%s)" % op
            else:
                rv = op()
        elif field_type:
            rv = self.represent(expression, field_type)
        elif isinstance(expression, (list, tuple)):
            rv = ",".join(self.represent(item, field_type) for item in expression)
        elif isinstance(expression, bool):
            rv = self.dialect.true_exp if expression else self.dialect.false_exp
        else:
            rv = expression
        return str(rv)

    def _expand_for_index(
        self, expression, field_type=None, colnames=False, query_env={}
    ):
        if isinstance(expression, Field):
            return expression._rname
        return self._expand(expression, field_type, colnames, query_env)

    @contextmanager
    def index_expander(self):
        self.expand = self._expand_for_index
        yield
        self.expand = self._expand

    def lastrowid(self, table):
        return self.cursor.lastrowid

    def _insert(self, table, fields):
        if fields:
            return self.dialect.insert(
                table._rname,
                ",".join(el[0]._rname for el in fields),
                ",".join(self.expand(v, f.type) for f, v in fields),
            )
        return self.dialect.insert_empty(table._rname)

    def insert(self, table, fields):
        query = self._insert(table, fields)
        try:
            self.execute(query)
        except:
            e = sys.exc_info()[1]
            if hasattr(table, "_on_insert_error"):
                return table._on_insert_error(table, fields, e)
            raise e
        if hasattr(table, "_primarykey"):
            pkdict = dict(
                [(k[0].name, k[1]) for k in fields if k[0].name in table._primarykey]
            )
            if pkdict:
                return pkdict
        id = self.lastrowid(table)
        if hasattr(table, "_primarykey") and len(table._primarykey) == 1:
            id = {table._primarykey[0]: id}
        if not isinstance(id, integer_types):
            return id
        rid = Reference(id)
        (rid._table, rid._record) = (table, None)
        return rid

    def _update(self, table, query, fields):
        sql_q = ""
        query_env = dict(current_scope=[table._tablename])
        if query:
            if use_common_filters(query):
                query = self.common_filter(query, [table])
            sql_q = self.expand(query, query_env=query_env)
        sql_v = ",".join(
            [
                "%s=%s"
                % (field._rname, self.expand(value, field.type, query_env=query_env))
                for (field, value) in fields
            ]
        )
        return self.dialect.update(table, sql_v, sql_q)

    def update(self, table, query, fields):
        sql = self._update(table, query, fields)
        try:
            self.execute(sql)
        except:
            e = sys.exc_info()[1]
            if hasattr(table, "_on_update_error"):
                return table._on_update_error(table, query, fields, e)
            raise e
        try:
            return self.cursor.rowcount
        except:
            return None

    def _delete(self, table, query):
        sql_q = ""
        query_env = dict(current_scope=[table._tablename])
        if query:
            if use_common_filters(query):
                query = self.common_filter(query, [table])
            sql_q = self.expand(query, query_env=query_env)
        return self.dialect.delete(table, sql_q)

    def delete(self, table, query):
        sql = self._delete(table, query)
        self.execute(sql)
        try:
            return self.cursor.rowcount
        except:
            return None

    def _colexpand(self, field, query_env):
        return self.expand(field, colnames=True, query_env=query_env)

    def _geoexpand(self, field, query_env):
        if (
            isinstance(field.type, str)
            and field.type.startswith("geo")
            and isinstance(field, Field)
        ):
            field = field.st_astext()
        return self.expand(field, query_env=query_env)

    def _build_joins_for_select(self, tablenames, param):
        if not isinstance(param, (tuple, list)):
            param = [param]
        tablemap = {}
        for item in param:
            if isinstance(item, Expression):
                item = item.first
            key = item._tablename
            if tablemap.get(key, item) is not item:
                raise ValueError("Name conflict in table list: %s" % key)
            tablemap[key] = item
        join_tables = [t._tablename for t in param if not isinstance(t, Expression)]
        join_on = [t for t in param if isinstance(t, Expression)]
        tables_to_merge = {}
        for t in join_on:
            tables_to_merge = merge_tablemaps(tables_to_merge, self.tables(t))
        join_on_tables = [t.first._tablename for t in join_on]
        for t in join_on_tables:
            if t in tables_to_merge:
                tables_to_merge.pop(t)
        important_tablenames = join_tables + join_on_tables + list(tables_to_merge)
        excluded = [t for t in tablenames if t not in important_tablenames]
        return (
            join_tables,
            join_on,
            tables_to_merge,
            join_on_tables,
            important_tablenames,
            excluded,
            tablemap,
        )

    def _select_wcols(
        self,
        query,
        fields,
        left=False,
        join=False,
        distinct=False,
        orderby=False,
        groupby=False,
        having=False,
        limitby=False,
        orderby_on_limitby=True,
        for_update=False,
        outer_scoped=[],
        required=None,
        cache=None,
        cacheable=None,
        processor=None,
        cte_collector=None,
    ):
        if cte_collector is None:
            cte_collector = dict(stack=[], seen=set(), is_recursive=False)
            is_toplevel = True
        else:
            is_toplevel = False

        #: parse tablemap
        tablemap = self.tables(query)
        #: apply common filters if needed
        if use_common_filters(query):
            query = self.common_filter(query, list(tablemap.values()))
        #: auto-adjust tables
        tablemap = merge_tablemaps(tablemap, self.tables(*fields))
        #: remove outer scoped tables if needed
        for item in outer_scoped:
            # FIXME: check for name conflicts
            tablemap.pop(item, None)
        if len(tablemap) < 1:
            raise SyntaxError("Set: no tables selected")

        query_tables = list(tablemap)
        #: check for_update argument
        # [Note - gi0baro] I think this should be removed since useless?
        #                  should affect only NoSQL?
        if self.can_select_for_update is False and for_update is True:
            raise SyntaxError("invalid select attribute: for_update")
        #: build joins (inner, left outer) and table names
        if join:
            (
                # FIXME? ijoin_tables is never used
                ijoin_tables,
                ijoin_on,
                itables_to_merge,
                ijoin_on_tables,
                iimportant_tablenames,
                iexcluded,
                itablemap,
            ) = self._build_joins_for_select(tablemap, join)
            tablemap = merge_tablemaps(tablemap, itables_to_merge)
            tablemap = merge_tablemaps(tablemap, itablemap)
        if left:
            (
                join_tables,
                join_on,
                tables_to_merge,
                join_on_tables,
                important_tablenames,
                excluded,
                jtablemap,
            ) = self._build_joins_for_select(tablemap, left)
            tablemap = merge_tablemaps(tablemap, tables_to_merge)
            tablemap = merge_tablemaps(tablemap, jtablemap)
        current_scope = outer_scoped + list(tablemap)
        query_env = dict(current_scope=current_scope, parent_scope=outer_scoped)
        #: prepare columns and expand fields
        colnames = [self._colexpand(x, query_env) for x in fields]
        sql_fields = ", ".join(self._geoexpand(x, query_env) for x in fields)
        table_alias = lambda name: tablemap[name].query_name(outer_scoped)[0]
        if join and not left:
            cross_joins = iexcluded + list(itables_to_merge)
            tokens = [table_alias(cross_joins[0])]
            tokens.extend(
                [
                    self.dialect.cross_join(table_alias(t), query_env)
                    for t in cross_joins[1:]
                ]
            )
            tokens.extend([self.dialect.join(t, query_env) for t in ijoin_on])
            sql_t = " ".join(tokens)
        elif not join and left:
            cross_joins = excluded + list(tables_to_merge)
            tokens = [table_alias(cross_joins[0])]
            tokens.extend(
                [
                    self.dialect.cross_join(table_alias(t), query_env)
                    for t in cross_joins[1:]
                ]
            )
            # FIXME: WTF? This is not correct syntax at least on PostgreSQL
            if join_tables:
                tokens.append(
                    self.dialect.left_join(
                        ",".join([table_alias(t) for t in join_tables]), query_env
                    )
                )
            tokens.extend([self.dialect.left_join(t, query_env) for t in join_on])
            sql_t = " ".join(tokens)
        elif join and left:
            all_tables_in_query = set(
                important_tablenames + iimportant_tablenames + query_tables
            )
            tables_in_joinon = set(join_on_tables + ijoin_on_tables)
            tables_not_in_joinon = list(
                all_tables_in_query.difference(tables_in_joinon)
            )
            tokens = [table_alias(tables_not_in_joinon[0])]
            tokens.extend(
                [
                    self.dialect.cross_join(table_alias(t), query_env)
                    for t in tables_not_in_joinon[1:]
                ]
            )
            tokens.extend([self.dialect.join(t, query_env) for t in ijoin_on])
            # FIXME: WTF? This is not correct syntax at least on PostgreSQL
            if join_tables:
                tokens.append(
                    self.dialect.left_join(
                        ",".join([table_alias(t) for t in join_tables]), query_env
                    )
                )
            tokens.extend([self.dialect.left_join(t, query_env) for t in join_on])
            sql_t = " ".join(tokens)
        else:
            sql_t = ", ".join(table_alias(t) for t in query_tables)
        #: expand query if needed
        if query:
            query = self.expand(query, query_env=query_env)
        if having:
            having = self.expand(having, query_env=query_env)
        #: groupby
        sql_grp = groupby
        if groupby:
            if isinstance(groupby, (list, tuple)):
                groupby = xorify(groupby)
            sql_grp = self.expand(groupby, query_env=query_env)
        #: orderby
        sql_ord = False
        if orderby:
            if isinstance(orderby, (list, tuple)):
                orderby = xorify(orderby)
            if str(orderby) == "<random>":
                sql_ord = self.dialect.random
            else:
                sql_ord = self.expand(orderby, query_env=query_env)
        #: set default orderby if missing
        if (
            limitby
            and not groupby
            and query_tables
            and orderby_on_limitby
            and not orderby
        ):
            sql_ord = ", ".join(
                [
                    tablemap[t][x].sqlsafe
                    for t in query_tables
                    if not isinstance(tablemap[t], Select)
                    for x in (getattr(tablemap[t], "_primarykey", None) or ["_id"])
                ]
            )

        #: build CTE
        [t.cte(cte_collector) for t in tablemap.values() if getattr(t, "is_cte", None)]
        if is_toplevel and cte_collector["stack"]:
            with_cte = [
                cte_collector["is_recursive"],
                ", ".join(cte_collector["stack"]),
            ]
        else:
            with_cte = None

        #: build sql using dialect
        return (
            colnames,
            self.dialect.select(
                sql_fields,
                sql_t,
                query,
                sql_grp,
                having,
                sql_ord,
                limitby,
                distinct,
                for_update and self.can_select_for_update,
                with_cte,
            ),
        )

    def _select(self, query, fields, attributes):
        return self._select_wcols(query, fields, **attributes)[1]

    def nested_select(self, query, fields, attributes):
        return Select(self.db, query, fields, attributes)

    def _select_aux_execute(self, sql):
        self.execute(sql)
        return self.cursor.fetchall()

    def _select_aux(self, sql, fields, attributes, colnames):
        cache = attributes.get("cache", None)
        if not cache:
            rows = self._select_aux_execute(sql)
        else:
            if isinstance(cache, dict):
                cache_model = cache["model"]
                time_expire = cache["expiration"]
                key = cache.get("key")
                if not key:
                    key = self.uri + "/" + sql + "/rows"
                    key = hashlib_md5(key).hexdigest()
            else:
                (cache_model, time_expire) = cache
                key = self.uri + "/" + sql + "/rows"
                key = hashlib_md5(key).hexdigest()
            rows = cache_model(
                key,
                lambda self=self, sql=sql: self._select_aux_execute(sql),
                time_expire,
            )
        if isinstance(rows, tuple):
            rows = list(rows)
        limitby = attributes.get("limitby", None) or (0,)
        rows = self.rowslice(rows, limitby[0], None)
        processor = attributes.get("processor", self.parse)
        cacheable = attributes.get("cacheable", False)
        return processor(rows, fields, colnames, cacheable=cacheable)

    def _cached_select(self, cache, sql, fields, attributes, colnames):
        del attributes["cache"]
        (cache_model, time_expire) = cache
        key = self.uri + "/" + sql
        key = hashlib_md5(key).hexdigest()
        args = (sql, fields, attributes, colnames)
        ret = cache_model(
            key, lambda self=self, args=args: self._select_aux(*args), time_expire
        )
        ret._restore_fields(fields)
        return ret

    def select(self, query, fields, attributes):
        colnames, sql = self._select_wcols(query, fields, **attributes)
        cache = attributes.get("cache", None)
        if cache and attributes.get("cacheable", False):
            return self._cached_select(cache, sql, fields, attributes, colnames)
        return self._select_aux(sql, fields, attributes, colnames)

    def iterselect(self, query, fields, attributes):
        colnames, sql = self._select_wcols(query, fields, **attributes)
        cacheable = attributes.get("cacheable", False)
        return self.iterparse(sql, fields, colnames, cacheable=cacheable)

    def _count(self, query, distinct=None):
        tablemap = self.tables(query)
        tablenames = list(tablemap)
        tables = list(tablemap.values())
        query_env = dict(current_scope=tablenames)
        sql_q = ""
        if query:
            if use_common_filters(query):
                query = self.common_filter(query, tables)
            sql_q = self.expand(query, query_env=query_env)
        sql_t = ",".join(self.table_alias(t, []) for t in tables)
        sql_fields = "*"
        if distinct:
            if isinstance(distinct, (list, tuple)):
                distinct = xorify(distinct)
            sql_fields = self.expand(distinct, query_env=query_env)
        return self.dialect.select(
            self.dialect.count(sql_fields, distinct), sql_t, sql_q
        )

    def count(self, query, distinct=None):
        self.execute(self._count(query, distinct))
        return self.cursor.fetchone()[0]

    def bulk_insert(self, table, items):
        return [self.insert(table, item) for item in items]

    def create_table(self, *args, **kwargs):
        return self.migrator.create_table(*args, **kwargs)

    def _drop_table_cleanup(self, table):
        super(SQLAdapter, self)._drop_table_cleanup(table)
        if table._dbt:
            self.migrator.file_delete(table._dbt)
            self.migrator.log("success!\n", table)

    def drop_table(self, table, mode=""):
        queries = self.dialect.drop_table(table, mode)
        for query in queries:
            if table._dbt:
                self.migrator.log(query + "\n", table)
            self.execute(query)
        self.commit()
        self._drop_table_cleanup(table)

    @deprecated("drop", "drop_table", "SQLAdapter")
    def drop(self, table, mode=""):
        return self.drop_table(table, mode="")

    def truncate(self, table, mode=""):
        # Prepare functions "write_to_logfile" and "close_logfile"
        try:
            queries = self.dialect.truncate(table, mode)
            for query in queries:
                self.migrator.log(query + "\n", table)
                self.execute(query)
            self.migrator.log("success!\n", table)
        finally:
            pass

    def create_index(self, table, index_name, *fields, **kwargs):
        expressions = [
            field._rname if isinstance(field, Field) else field for field in fields
        ]
        sql = self.dialect.create_index(index_name, table, expressions, **kwargs)
        try:
            self.execute(sql)
            self.commit()
        except Exception as e:
            self.rollback()
            err = (
                "Error creating index %s\n  Driver error: %s\n"
                + "  SQL instruction: %s"
            )
            raise RuntimeError(err % (index_name, str(e), sql))
        return True

    def drop_index(self, table, index_name, if_exists=False):
        sql = self.dialect.drop_index(index_name, table, if_exists)
        try:
            self.execute(sql)
            self.commit()
        except Exception as e:
            self.rollback()
            err = "Error dropping index %s\n  Driver error: %s"
            raise RuntimeError(err % (index_name, str(e)))
        return True

    def distributed_transaction_begin(self, key):
        pass

    @with_connection
    def commit(self):
        return self.connection.commit()

    @with_connection
    def rollback(self):
        return self.connection.rollback()

    @with_connection
    def prepare(self, key):
        self.connection.prepare()

    @with_connection
    def commit_prepared(self, key):
        self.connection.commit()

    @with_connection
    def rollback_prepared(self, key):
        self.connection.rollback()

    def create_sequence_and_triggers(self, query, table, **args):
        self.execute(query)

    def sqlsafe_table(self, tablename, original_tablename=None):
        if original_tablename is not None:
            return self.dialect.alias(original_tablename, tablename)
        return self.dialect.quote(tablename)

    def sqlsafe_field(self, fieldname):
        return self.dialect.quote(fieldname)

    def table_alias(self, tbl, current_scope=[]):
        if isinstance(tbl, basestring):
            tbl = self.db[tbl]
        return tbl.query_name(current_scope)[0]

    def id_query(self, table):
        pkeys = getattr(table, "_primarykey", None)
        if pkeys:
            return table[pkeys[0]] != None
        return table._id != None


class NoSQLAdapter(BaseAdapter):
    can_select_for_update = False

    def commit(self):
        pass

    def rollback(self):
        pass

    def prepare(self):
        pass

    def commit_prepared(self, key):
        pass

    def rollback_prepared(self, key):
        pass

    def id_query(self, table):
        return table._id > 0

    def create_table(self, table, migrate=True, fake_migrate=False, polymodel=None):
        table._dbt = None
        table._notnulls = []
        for field_name in table.fields:
            if table[field_name].notnull:
                table._notnulls.append(field_name)
        table._uniques = []
        for field_name in table.fields:
            if table[field_name].unique:
                # this is unnecessary if the fields are indexed and unique
                table._uniques.append(field_name)

    def drop_table(self, table, mode=""):
        ctable = self.connection[table._tablename]
        ctable.drop()
        self._drop_table_cleanup(table)

    @deprecated("drop", "drop_table", "SQLAdapter")
    def drop(self, table, mode=""):
        return self.drop_table(table, mode="")

    def _select(self, *args, **kwargs):
        raise NotOnNOSQLError("Nested queries are not supported on NoSQL databases")

    def nested_select(self, *args, **kwargs):
        raise NotOnNOSQLError("Nested queries are not supported on NoSQL databases")


class NullAdapter(BaseAdapter):
    def _load_dependencies(self):
        from ..dialects.base import CommonDialect

        self.dialect = CommonDialect(self)

    def find_driver(self):
        pass

    def connector(self):
        return NullDriver()



import os.path
import re

from .._compat import PY2, iterkeys, long, to_unicode, with_metaclass
from .._globals import IDENTITY, THREAD_LOCAL
from ..drivers import psycopg2_adapt
from ..utils import split_uri_args
from . import AdapterMeta, adapters, with_connection, with_connection_or_raise
from .base import SQLAdapter


class PostgreMeta(AdapterMeta):
    def __call__(cls, *args, **kwargs):
        if cls not in [Postgre, PostgreNew, PostgreBoolean]:
            return AdapterMeta.__call__(cls, *args, **kwargs)
        # choose driver according uri
        available_drivers = [
            driver
            for driver in cls.drivers
            if driver in iterkeys(kwargs["db"]._drivers_available)
        ]
        uri_items = kwargs["uri"].split("://", 1)[0].split(":")
        uri_driver = uri_items[1] if len(uri_items) > 1 else None
        if uri_driver and uri_driver in available_drivers:
            driver = uri_driver
        else:
            driver = available_drivers[0] if available_drivers else cls.drivers[0]
        cls = adapters._registry_[uri_items[0] + ":" + driver]
        return AdapterMeta.__call__(cls, *args, **kwargs)


@adapters.register_for("postgres")
class Postgre(with_metaclass(PostgreMeta, SQLAdapter)):
    dbengine = "postgres"
    drivers = ("psycopg2",)
    support_distributed_transaction = True

    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]*|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)"
        r"(\?(?P<uriargs>.*))?$"
    )  # sslmode, ssl (no value) and unix_socket

    def __init__(
        self,
        db,
        uri,
        pool_size=0,
        folder=None,
        db_codec="UTF-8",
        credential_decoder=IDENTITY,
        driver_args={},
        adapter_args={},
        srid=4326,
        after_connection=None,
    ):
        self.srid = srid
        super(Postgre, self).__init__(
            db,
            uri,
            pool_size,
            folder,
            db_codec,
            credential_decoder,
            driver_args,
            adapter_args,
            after_connection,
        )
        self._config_json()

    def _initialize_(self):
        super(Postgre, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        host = m.group("host")
        uriargs = m.group("uriargs")
        if uriargs:
            uri_args = split_uri_args(uriargs, need_equal=False)
        else:
            uri_args = dict()
        socket = uri_args.get("unix_socket")
        if not host and not socket:
            raise SyntaxError("Host or UNIX socket name required")
        db = m.group("db")
        self.driver_args.update(user=user, database=db)
        if password is not None:
            self.driver_args["password"] = password
        if socket:
            if not os.path.exists(socket):
                raise ValueError("UNIX socket %r not found" % socket)
            if self.driver_name == "psycopg2":
                # the psycopg2 driver let you configure the socket directory
                # only (not the socket file name) by passing it as the host
                # (must be an absolute path otherwise the driver tries a TCP/IP
                # connection to host); this behaviour is due to the underlying
                # libpq used by the driver
                socket_dir = os.path.abspath(os.path.dirname(socket))
                self.driver_args["host"] = socket_dir
        else:
            port = int(m.group("port") or 5432)
            self.driver_args.update(host=host, port=port)
            sslmode = uri_args.get("sslmode")
            if sslmode and self.driver_name == "psycopg2":
                self.driver_args["sslmode"] = sslmode
        if self.driver:
            self.__version__ = "%s %s" % (self.driver.__name__, self.driver.__version__)
        else:
            self.__version__ = None
        THREAD_LOCAL._pydal_last_insert_ = None
        self.get_connection()

    def _get_json_dialect(self):
        from ..dialects.postgre import PostgreDialectJSON

        return PostgreDialectJSON

    def _get_json_parser(self):
        from ..parsers.postgre import PostgreAutoJSONParser

        return PostgreAutoJSONParser

    @property
    def _last_insert(self):
        return THREAD_LOCAL._pydal_last_insert_

    @_last_insert.setter
    def _last_insert(self, value):
        THREAD_LOCAL._pydal_last_insert_ = value

    def connector(self):
        return self.driver.connect(**self.driver_args)

    def after_connection(self):
        self.execute("SET CLIENT_ENCODING TO 'UTF8'")
        self.execute("SET standard_conforming_strings=on;")

    def lastrowid(self, table):
        if self._last_insert:
            return long(self.cursor.fetchone()[0])
        sequence_name = table._sequence_name
        self.execute("SELECT currval(%s);" % self.adapt(sequence_name))
        return long(self.cursor.fetchone()[0])

    def _insert(self, table, fields):
        self._last_insert = None
        if fields:
            retval = None
            if hasattr(table, "_id"):
                self._last_insert = (table._id, 1)
                retval = table._id._rname
            return self.dialect.insert(
                table._rname,
                ",".join(el[0]._rname for el in fields),
                ",".join(self.expand(v, f.type) for f, v in fields),
                retval,
            )
        return self.dialect.insert_empty(table._rname)

    @with_connection
    def prepare(self, key):
        self.execute("PREPARE TRANSACTION '%s';" % key)

    @with_connection
    def commit_prepared(self, key):
        self.execute("COMMIT PREPARED '%s';" % key)

    @with_connection
    def rollback_prepared(self, key):
        self.execute("ROLLBACK PREPARED '%s';" % key)


@adapters.register_for("postgres:psycopg2")
class PostgrePsyco(Postgre):
    drivers = ("psycopg2",)

    def _config_json(self):
        use_json = (
            self.driver.__version__ >= "2.0.12"
            and self.connection.server_version >= 90200
        )
        if use_json:
            self.dialect = self._get_json_dialect()(self)
            if self.driver.__version__ >= "2.5.0":
                self.parser = self._get_json_parser()(self)

    def adapt(self, obj):
        adapted = psycopg2_adapt(obj)
        # deal with new relic Connection Wrapper (newrelic>=2.10.0.8)
        cxn = getattr(self.connection, "__wrapped__", self.connection)
        adapted.prepare(cxn)
        rv = adapted.getquoted()
        if not PY2:
            if isinstance(rv, bytes):
                return rv.decode("utf-8")
        return rv


@adapters.register_for("postgres2")
class PostgreNew(Postgre):
    def _get_json_dialect(self):
        from ..dialects.postgre import PostgreDialectArraysJSON

        return PostgreDialectArraysJSON

    def _get_json_parser(self):
        from ..parsers.postgre import PostgreNewAutoJSONParser

        return PostgreNewAutoJSONParser


@adapters.register_for("postgres2:psycopg2")
class PostgrePsycoNew(PostgrePsyco, PostgreNew):
    pass


@adapters.register_for("postgres3")
class PostgreBoolean(PostgreNew):
    def _get_json_dialect(self):
        from ..dialects.postgre import PostgreDialectBooleanJSON

        return PostgreDialectBooleanJSON

    def _get_json_parser(self):
        from ..parsers.postgre import PostgreBooleanAutoJSONParser

        return PostgreBooleanAutoJSONParser


@adapters.register_for("postgres3:psycopg2")
class PostgrePsycoBoolean(PostgrePsycoNew, PostgreBoolean):
    pass


@adapters.register_for("jdbc:postgres")
class JDBCPostgre(Postgre):
    drivers = ("zxJDBC",)

    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)$"
    )

    def _initialize_(self):
        super(Postgre, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        if password is None:
            password = ""
        host = m.group("host")
        db = m.group("db")
        port = m.group("port") or "5432"
        self.dsn = ("jdbc:postgresql://%s:%s/%s" % (host, port, db), user, password)
        if self.driver:
            self.__version__ = "%s %s" % (self.driver.__name__, self.driver.__version__)
        else:
            self.__version__ = None
        THREAD_LOCAL._pydal_last_insert_ = None
        self.get_connection()

    def connector(self):
        return self.driver.connect(*self.dsn, **self.driver_args)

    def after_connection(self):
        self.connection.set_client_encoding("UTF8")
        self.execute("BEGIN;")
        self.execute("SET CLIENT_ENCODING TO 'UNICODE';")

    def _config_json(self):
        use_json = self.connection.dbversion >= "9.2.0"
        if use_json:
            self.dialect = self._get_json_dialect()(self)



import unittest

from pydal import DAL, Field
from pydal.validators import IS_NOT_IN_DB
from pydal.restapi import RestAPI, ALLOW_ALL_POLICY, DENY_ALL_POLICY, __version__


class TestRestAPI(unittest.TestCase):
    def setUp(self):
        db = DAL("sqlite:memory")

        db.define_table("color", Field("name", requires=IS_NOT_IN_DB(db, "color.name")))
        db.color.insert(name="red")
        db.color.insert(name="green")
        db.color.insert(name="blue")

        db.define_table("thing", Field("name"), Field("color", "reference color"))
        db.thing.insert(name="Chair", color=1)
        db.thing.insert(name="Chair", color=2)
        db.thing.insert(name="Table", color=1)
        db.thing.insert(name="Table", color=3)
        db.thing.insert(name="Lamp", color=2)

        db.define_table(
            "rel",
            Field("a", "reference thing"),
            Field("desc"),
            Field("b", "reference thing"),
        )
        db.rel.insert(a=1, b=2, desc="is like")
        db.rel.insert(a=3, b=4, desc="is like")
        db.rel.insert(a=1, b=3, desc="is under")
        db.rel.insert(a=2, b=4, desc="is under")
        db.rel.insert(a=5, b=4, desc="is above")

        api = RestAPI(db, ALLOW_ALL_POLICY)

        self.db = db
        self.api = api

    def test_search(self):
        api = self.api
        api.policy = ALLOW_ALL_POLICY

        self.assertEqual(
            api.search("rel", {"@limit": 0, "@model": "true"}),
            {
                "items": [],
                "count": 5,
                "model": [
                    {
                        "name": "id",
                        "label": "Id",
                        "default": None,
                        "type": "id",
                        "regex": "[1-9]\\d*",
                        "required": False,
                        "unique": False,
                        "post_writable": True,
                        "put_writable": True,
                        "options": None,
                        "referenced_by": [],
                    },
                    {
                        "name": "a",
                        "label": "A",
                        "default": None,
                        "type": "reference",
                        "references": "thing",
                        "regex": None,
                        "required": False,
                        "unique": False,
                        "post_writable": True,
                        "put_writable": True,
                        "options": None,
                    },
                    {
                        "name": "desc",
                        "label": "Desc",
                        "default": None,
                        "type": "string",
                        "regex": None,
                        "required": False,
                        "unique": False,
                        "post_writable": True,
                        "put_writable": True,
                        "options": None,
                    },
                    {
                        "name": "b",
                        "label": "B",
                        "default": None,
                        "type": "reference",
                        "references": "thing",
                        "regex": None,
                        "required": False,
                        "unique": False,
                        "post_writable": True,
                        "put_writable": True,
                        "options": None,
                    },
                ],
            },
        )
        self.assertEqual(
            api.search("color", {"name.eq": "red"}),
            {"count": 1, "items": [{"id": 1, "name": "red"}]},
        )
        self.assertEqual(
            api.search("thing", {"name.eq": "Chair"}),
            {
                "count": 2,
                "items": [
                    {"name": "Chair", "color": 1, "id": 1},
                    {"name": "Chair", "color": 2, "id": 2},
                ],
            },
        )
        self.assertEqual(
            api.search("rel[a,b]", {"desc.eq": "is like"}),
            {"count": 2, "items": [{"b": 2, "a": 1}, {"b": 4, "a": 3}]},
        )
        self.assertEqual(
            api.search("thing[name]", {"color.name.eq": "red"}),
            {"count": 2, "items": [{"name": "Chair"}, {"name": "Table"}]},
        )
        self.assertEqual(
            api.search("thing[name]", {"not.color.name.eq": "red"}),
            {
                "count": 3,
                "items": [{"name": "Chair"}, {"name": "Table"}, {"name": "Lamp"}],
            },
        )
        self.assertEqual(
            api.search("thing[name]", {"a.rel.desc": "is above"}),
            {"count": 1, "items": [{"name": "Lamp"}]},
        )
        self.assertEqual(
            api.search("thing[name]", {"a.rel.b.name": "Table"}),
            {
                "count": 4,
                "items": [
                    {"name": "Chair"},
                    {"name": "Chair"},
                    {"name": "Table"},
                    {"name": "Lamp"},
                ],
            },
        )
        self.assertEqual(
            api.search(
                "thing[name]", {"a.rel.b.name": "Table", "a.rel.desc": "is above"}
            ),
            {"count": 1, "items": [{"name": "Lamp"}]},
        )
        self.assertEqual(
            api.search("thing", {"@lookup": "color"}),
            {
                "count": 5,
                "items": [
                    {"name": "Chair", "color": {"name": "red", "id": 1}, "id": 1},
                    {"name": "Chair", "color": {"name": "green", "id": 2}, "id": 2},
                    {"name": "Table", "color": {"name": "red", "id": 1}, "id": 3},
                    {"name": "Table", "color": {"name": "blue", "id": 3}, "id": 4},
                    {"name": "Lamp", "color": {"name": "green", "id": 2}, "id": 5},
                ],
            },
        )
        self.assertEqual(
            api.search("thing", {"@lookup": "color[name]"}),
            {
                "count": 5,
                "items": [
                    {"name": "Chair", "color": {"name": "red"}, "id": 1},
                    {"name": "Chair", "color": {"name": "green"}, "id": 2},
                    {"name": "Table", "color": {"name": "red"}, "id": 3},
                    {"name": "Table", "color": {"name": "blue"}, "id": 4},
                    {"name": "Lamp", "color": {"name": "green"}, "id": 5},
                ],
            },
        )
        self.assertEqual(
            api.search("thing", {"@lookup": "color!:color[name]"}),
            {
                "count": 5,
                "items": [
                    {"name": "Chair", "color.name": "red", "id": 1},
                    {"name": "Chair", "color.name": "green", "id": 2},
                    {"name": "Table", "color.name": "red", "id": 3},
                    {"name": "Table", "color.name": "blue", "id": 4},
                    {"name": "Lamp", "color.name": "green", "id": 5},
                ],
            },
        )
        self.assertEqual(
            api.search("thing", {"@lookup": "related:a.rel[desc]"}),
            {
                "count": 5,
                "items": [
                    {
                        "name": "Chair",
                        "related": [{"desc": "is like"}, {"desc": "is under"}],
                        "color": 1,
                        "id": 1,
                    },
                    {
                        "name": "Chair",
                        "related": [{"desc": "is under"}],
                        "color": 2,
                        "id": 2,
                    },
                    {
                        "name": "Table",
                        "related": [{"desc": "is like"}],
                        "color": 1,
                        "id": 3,
                    },
                    {"name": "Table", "related": [], "color": 3, "id": 4},
                    {
                        "name": "Lamp",
                        "related": [{"desc": "is above"}],
                        "color": 2,
                        "id": 5,
                    },
                ],
            },
        )
        self.assertEqual(
            api.search("thing", {"@lookup": "related:a.rel[desc].b[name]"}),
            {
                "count": 5,
                "items": [
                    {
                        "name": "Chair",
                        "related": [
                            {"b": {"name": "Chair"}, "desc": "is like"},
                            {"b": {"name": "Table"}, "desc": "is under"},
                        ],
                        "color": 1,
                        "id": 1,
                    },
                    {
                        "name": "Chair",
                        "related": [{"b": {"name": "Table"}, "desc": "is under"}],
                        "color": 2,
                        "id": 2,
                    },
                    {
                        "name": "Table",
                        "related": [{"b": {"name": "Table"}, "desc": "is like"}],
                        "color": 1,
                        "id": 3,
                    },
                    {"name": "Table", "related": [], "color": 3, "id": 4},
                    {
                        "name": "Lamp",
                        "related": [{"b": {"name": "Table"}, "desc": "is above"}],
                        "color": 2,
                        "id": 5,
                    },
                ],
            },
        )
        self.assertEqual(
            api.search(
                "thing",
                {
                    "@lookup": "color[name],related:a.rel[desc].b[name]",
                    "@offset": 1,
                    "@limit": 2,
                },
            ),
            {
                "items": [
                    {
                        "name": "Chair",
                        "related": [{"b": {"name": "Table"}, "desc": "is under"}],
                        "color": {"name": "green"},
                        "id": 2,
                    },
                    {
                        "name": "Table",
                        "related": [{"b": {"name": "Table"}, "desc": "is like"}],
                        "color": {"name": "red"},
                        "id": 3,
                    },
                ]
            },
        )
        self.assertEqual(
            api.search(
                "thing",
                {
                    "@lookup": "color[name],related!:a.rel[desc].b[name]",
                    "@offset": 1,
                    "@limit": 2,
                },
            ),
            {
                "items": [
                    {
                        "name": "Chair",
                        "related": [{"name": "Table", "desc": "is under"}],
                        "color": {"name": "green"},
                        "id": 2,
                    },
                    {
                        "name": "Table",
                        "related": [{"name": "Table", "desc": "is like"}],
                        "color": {"name": "red"},
                        "id": 3,
                    },
                ]
            },
        )
        self.assertEqual(
            api.search("color", {"name.contains": "ee"}),
            {"count": 1, "items": [{"id": 2, "name": "green"}]},
        )
        self.assertEqual(
            api.search("color", {"name.in": "blue,green"}),
            {
                "count": 2,
                "items": [{"id": 2, "name": "green"}, {"id": 3, "name": "blue"}],
            },
        )

    def test_REST(self):

        api = self.api
        api.policy = ALLOW_ALL_POLICY

        response = api("GET", "color", None, {"name.eq": "red"})
        del response["timestamp"]
        self.assertEqual(
            response,
            {
                "count": 1,
                "status": "success",
                "code": 200,
                "items": [{"id": 1, "name": "red"}],
                "api_version": __version__,
            },
        )
        response = api("POST", "color", post_vars={"name": "magenta"})
        del response["timestamp"]
        self.assertEqual(
            response,
            {
                "status": "success",
                "errors": {},
                "code": 200,
                "id": 4,
                "api_version": __version__,
            },
        )
        response = api("POST", "color", post_vars={"name": "magenta"})
        del response["timestamp"]
        self.assertEqual(
            response,
            {
                "status": "error",
                "errors": {"name": "Value already in database or empty"},
                "code": 422,
                "message": "Validation Errors",
                "id": None,
                "api_version": __version__,
            },
        )
        response = api("PUT", "color", 4, post_vars={"name": "Magenta"})
        del response["timestamp"]
        self.assertEqual(
            response,
            {
                "status": "success",
                "updated": 1,
                "errors": {},
                "code": 200,
                "api_version": "0.1",
                "id": 4,
            },
        )
        response = api("DELETE", "color", 4)
        del response["timestamp"]
        self.assertEqual(
            response,
            {"deleted": 1, "status": "success", "code": 200, "api_version": "0.1"},
        )

    def test_policies(self):

        api = self.api
        api.policy = DENY_ALL_POLICY

        response = api("GET", "color", None, {"name.eq": "red"})
        del response["timestamp"]
        self.assertEqual(
            response,
            {
                "status": "error",
                "message": "No policy for this object",
                "code": 401,
                "api_version": __version__,
            },
        )



import copy
import datetime
import locale
import os
import pickle
import sys

from ._compat import PY2, exists, iteritems, pjoin, string_types, to_bytes
from ._load import portalocker
from .helpers.classes import DatabaseStoredFile, SQLCustomType


class Migrator(object):
    def __init__(self, adapter):
        self.adapter = adapter

    @property
    def db(self):
        return self.adapter.db

    @property
    def dialect(self):
        return self.adapter.dialect

    @property
    def dbengine(self):
        return self.adapter.dbengine

    def create_table(self, table, migrate=True, fake_migrate=False, polymodel=None):
        db = table._db
        table._migrate = migrate
        fields = []
        reffs = str()
        # PostGIS geo fields are added after the table has been created
        postcreation_fields = []
        sql_fields = {}
        sql_fields_aux = {}
        TFK = {}
        tablename = table._tablename
        types = self.adapter.types
        for sortable, field in enumerate(table, start=1):
            if self.db._ignore_field_case:
                field_name = field.name.lower()
                field_rname = field._rname.lower()
            else:
                field_name = field.name
                field_rname = field._rname
            if self.dbengine == "oracle":
                # Oracle needs all field names quoted to ensure consistent case
                field_rname = self.dialect.quote(field_rname)
            field_type = field.type
            if isinstance(field_type, SQLCustomType):
                ftype = field_type.native or field_type.type
            elif field_type.startswith(("reference", "big-reference")):
                if field_type.startswith("reference"):
                    referenced = field_type[10:].strip()
                    type_name = "reference"
                else:
                    referenced = field_type[14:].strip()
                    type_name = "big-reference"

                if referenced == ".":
                    referenced = tablename
                constraint_name = self.dialect.constraint_name(
                    table._raw_rname, field._raw_rname
                )
                # if not '.' in referenced \
                #         and referenced != tablename \
                #         and hasattr(table,'_primarykey'):
                #     ftype = types['integer']
                # else:
                try:
                    rtable = db[referenced]
                    rfield = rtable._id
                    rfieldname = rfield.name
                    rtablename = referenced
                except (KeyError, ValueError, AttributeError) as e:
                    self.db.logger.debug("Error: %s" % e)
                    try:
                        rtablename, rfieldname = referenced.split(".")
                        rtable = db[rtablename]
                        rfield = rtable[rfieldname]
                    except Exception as e:
                        self.db.logger.debug("Error: %s" % e)
                        raise KeyError(
                            "Cannot resolve reference %s in %s definition"
                            % (referenced, table._tablename)
                        )

                # must be PK reference or unique
                rfield_rname = rfield._rname
                if self.dbengine == "oracle":
                    rfield_rname = self.dialect.quote(rfield_rname)

                if (
                    not rfield.type.startswith(("reference", "big-reference"))
                    and getattr(rtable, "_primarykey", None)
                    and rfieldname in rtable._primarykey
                    or rfield.unique
                ):
                    ftype = types[rfield.type[:9]] % dict(length=rfield.length)
                    # multicolumn primary key reference?
                    if not rfield.unique and len(rtable._primarykey) > 1:
                        # then it has to be a table level FK
                        if rtablename not in TFK:
                            TFK[rtablename] = {}
                        TFK[rtablename][rfieldname] = field_name
                    else:
                        fk = rtable._rname + " (" + rfield._rname + ")"
                        if self.dbengine == "oracle":
                            fk = (
                                self.dialect.quote(rtable._rname)
                                + " ("
                                + rfield_rname
                                + ")"
                            )
                        reffs = reffs + types["reference FK"] % dict(
                            # should be quoted
                            constraint_name=constraint_name,
                            foreign_key=fk,
                            table_name=table._rname,
                            field_name=field._rname,
                            on_delete_action=field.ondelete,
                            on_update_action=field.onupdate,
                        )
                else:
                    # make a guess here for circular references
                    if referenced in db:
                        id_fieldname = db[referenced]._id._rname
                    elif referenced == tablename:
                        id_fieldname = table._id._rname
                    else:  # make a guess
                        id_fieldname = self.dialect.quote("id")
                    # gotcha: the referenced table must be defined before
                    # the referencing one to be able to create the table
                    # Also if it's not recommended, we can still support
                    # references to tablenames without rname to make
                    # migrations and model relationship work also if tables
                    # are not defined in order
                    if referenced == tablename:
                        real_referenced = db[referenced]._rname
                    else:
                        real_referenced = (
                            referenced in db and db[referenced]._rname or referenced
                        )
                    if self.dbengine == "oracle":
                        real_referenced = self.dialect.quote(real_referenced)
                    rfield = db[referenced]._id
                    ftype_info = dict(
                        index_name=self.dialect.quote(field._raw_rname + "__idx"),
                        field_name=field_rname,
                        constraint_name=self.dialect.quote(constraint_name),
                        foreign_key="%s (%s)" % (real_referenced, rfield_rname),
                        on_delete_action=field.ondelete,
                        on_update_action=field.onupdate,
                    )
                    ftype_info["null"] = (
                        " NOT NULL" if field.notnull else self.dialect.allow_null
                    )
                    ftype_info["unique"] = " UNIQUE" if field.unique else ""
                    ftype = types[type_name] % ftype_info
            elif field_type.startswith("list:reference"):
                ftype = types[field_type[:14]]
            elif field_type.startswith("decimal"):
                precision, scale = map(int, field_type[8:-1].split(","))
                ftype = types[field_type[:7]] % dict(precision=precision, scale=scale)
            elif field_type.startswith("geo"):
                if not hasattr(self.adapter, "srid"):
                    raise RuntimeError("Adapter does not support geometry")
                srid = self.adapter.srid
                geotype, parms = field_type[:-1].split("(")
                if geotype not in types:
                    raise SyntaxError(
                        "Field: unknown field type: %s for %s"
                        % (field_type, field_name)
                    )
                ftype = types[geotype]
                if self.dbengine == "postgres" and geotype == "geometry":
                    if self.db._ignore_field_case is True:
                        field_name = field_name.lower()
                    # parameters: schema, srid, dimension
                    dimension = 2  # GIS.dimension ???
                    parms = parms.split(",")
                    if len(parms) == 3:
                        schema, srid, dimension = parms
                    elif len(parms) == 2:
                        schema, srid = parms
                    else:
                        schema = parms[0]
                    ftype = (
                        "SELECT AddGeometryColumn ('%%(schema)s', '%%(tablename)s', '%%(fieldname)s', %%(srid)s, '%s', %%(dimension)s);"
                        % types[geotype]
                    )
                    ftype = ftype % dict(
                        schema=schema,
                        tablename=table._raw_rname,
                        fieldname=field._raw_rname,
                        srid=srid,
                        dimension=dimension,
                    )
                    postcreation_fields.append(ftype)
            elif field_type not in types:
                raise SyntaxError(
                    "Field: unknown field type: %s for %s" % (field_type, field_name)
                )
            else:
                ftype = types[field_type] % {"length": field.length}

            if not field_type.startswith(("id", "reference", "big-reference")):
                if field.notnull:
                    ftype += " NOT NULL"
                else:
                    ftype += self.dialect.allow_null
                if field.unique:
                    ftype += " UNIQUE"
                if field.custom_qualifier:
                    ftype += " %s" % field.custom_qualifier

            # add to list of fields
            sql_fields[field_name] = dict(
                length=field.length,
                unique=field.unique,
                notnull=field.notnull,
                sortable=sortable,
                type=str(field_type),
                sql=ftype,
                rname=field_rname,
                raw_rname=field._raw_rname,
            )

            if field.notnull and field.default is not None:
                # Caveat: sql_fields and sql_fields_aux
                # differ for default values.
                # sql_fields is used to trigger migrations and sql_fields_aux
                # is used for create tables.
                # The reason is that we do not want to trigger
                # a migration simply because a default value changes.
                not_null = self.dialect.not_null(field.default, field_type)
                ftype = ftype.replace("NOT NULL", not_null)
            sql_fields_aux[field_name] = dict(sql=ftype)
            # Postgres - PostGIS:
            # geometry fields are added after the table has been created, not now
            if not (self.dbengine == "postgres" and field_type.startswith("geom")):
                fields.append("%s %s" % (field_rname, ftype))
        other = ";"

        # backend-specific extensions to fields
        if self.dbengine == "mysql":
            if not hasattr(table, "_primarykey"):
                fields.append("PRIMARY KEY (%s)" % (table._id._rname))
            engine = self.adapter.adapter_args.get("engine", "InnoDB")
            other = " ENGINE=%s CHARACTER SET utf8;" % engine

        fields = ",\n    ".join(fields) + reffs
        for rtablename in TFK:
            rtable = db[rtablename]
            rfields = TFK[rtablename]
            pkeys = [rtable[pk]._rname for pk in rtable._primarykey]
            fk_fields = [table[rfields[k]] for k in rtable._primarykey]
            fkeys = [f._rname for f in fk_fields]
            constraint_name = self.dialect.constraint_name(
                table._raw_rname, "_".join(f._raw_rname for f in fk_fields)
            )
            on_delete = list(set(f.ondelete for f in fk_fields))
            on_update = list(set(f.onupdate for f in fk_fields))
            if len(on_delete) > 1:
                raise SyntaxError(
                    "Table %s has incompatible ON DELETE actions in multi-field foreign key."
                    % table._dalname
                )
            if len(on_update) > 1:
                raise SyntaxError(
                    "Table %s has incompatible ON UPDATE actions in multi-field foreign key."
                    % table._dalname
                )
            tfk_field_name = ", ".join(fkeys)
            tfk_foreign_key = ", ".join(pkeys)
            tfk_foreign_table = rtable._rname
            if self.dbengine == "oracle":
                tfk_field_name = ", ".join([self.dialect.quote(fkey) for fkey in fkeys])
                tfk_foreign_key = ", ".join(
                    [self.dialect.quote(pkey) for pkey in pkeys]
                )
                tfk_foreign_table = self.dialect.quote(rtable._rname)
            fields = (
                fields
                + ",\n    "
                + types["reference TFK"]
                % dict(
                    constraint_name=constraint_name,
                    table_name=table._rname,
                    field_name=tfk_field_name,
                    foreign_table=tfk_foreign_table,
                    foreign_key=tfk_foreign_key,
                    on_delete_action=on_delete[0],
                    on_update_action=on_update[0],
                )
            )

        table_rname = table._rname
        if self.dbengine == "oracle":
            # must be explicitly quoted to preserve case
            table_rname = self.dialect.quote(table_rname)

        if getattr(table, "_primarykey", None):
            query = "CREATE TABLE %s(\n    %s,\n    %s) %s" % (
                table_rname,
                fields,
                self.dialect.primary_key(
                    ", ".join([table[pk]._rname for pk in table._primarykey])
                ),
                other,
            )
        else:
            query = "CREATE TABLE %s(\n    %s\n)%s" % (table_rname, fields, other)

        uri = self.adapter.uri
        if uri.startswith("sqlite:///") or uri.startswith("spatialite:///"):
            if PY2:
                path_encoding = (
                    sys.getfilesystemencoding()
                    or locale.getdefaultlocale()[1]
                    or "utf8"
                )
                dbpath = uri[9 : uri.rfind("/")].decode("utf8").encode(path_encoding)
            else:
                dbpath = uri[9 : uri.rfind("/")]
        else:
            dbpath = self.adapter.folder

        if not migrate:
            return query
        elif uri.startswith("sqlite:memory") or uri.startswith("spatialite:memory"):
            table._dbt = None
        elif isinstance(migrate, string_types):
            table._dbt = pjoin(dbpath, migrate)
        else:
            table._dbt = pjoin(dbpath, "%s_%s.table" % (db._uri_hash, tablename))

        if not table._dbt or not self.file_exists(table._dbt):
            if table._dbt:
                self.log(
                    "timestamp: %s\n%s\n"
                    % (datetime.datetime.today().isoformat(), query),
                    table,
                )
            if not fake_migrate:
                self.adapter.create_sequence_and_triggers(query, table)
                db.commit()
                # Postgres geom fields are added now,
                # after the table has been created
                for query in postcreation_fields:
                    self.adapter.execute(query)
                    db.commit()
            if table._dbt:
                tfile = self.file_open(table._dbt, "wb")
                pickle.dump(sql_fields, tfile)
                self.file_close(tfile)
                if fake_migrate:
                    self.log("faked!\n", table)
                else:
                    self.log("success!\n", table)
        else:
            tfile = self.file_open(table._dbt, "rb")
            try:
                sql_fields_old = pickle.load(tfile)
            except EOFError:
                self.file_close(tfile)
                raise RuntimeError("File %s appears corrupted" % table._dbt)
            self.file_close(tfile)
            # add missing rnames
            for key, item in sql_fields_old.items():
                tmp = sql_fields.get(key)
                if tmp:
                    item.setdefault("rname", tmp["rname"])
                    item.setdefault("raw_rname", tmp["raw_rname"])
                else:
                    item.setdefault("rname", self.dialect.quote(key))
                    item.setdefault("raw_rname", key)
            if sql_fields != sql_fields_old:
                self.migrate_table(
                    table,
                    sql_fields,
                    sql_fields_old,
                    sql_fields_aux,
                    None,
                    fake_migrate=fake_migrate,
                )
        return query

    def _fix(self, item):
        k, v = item
        if self.dbengine == "oracle" and "rname" in v:
            v["rname"] = self.dialect.quote(v["rname"])
        if not isinstance(v, dict):
            v = dict(type="unknown", sql=v)
        if self.db._ignore_field_case is not True:
            return k, v
        return k.lower(), v

    def migrate_table(
        self,
        table,
        sql_fields,
        sql_fields_old,
        sql_fields_aux,
        logfile,
        fake_migrate=False,
    ):
        # logfile is deprecated (moved to adapter.log method)
        db = table._db
        db._migrated.append(table._tablename)
        tablename = table._tablename
        if self.dbengine in ("firebird",):
            drop_expr = "ALTER TABLE %s DROP %s;"
        else:
            drop_expr = "ALTER TABLE %s DROP COLUMN %s;"
        field_types = dict(
            (x.lower(), table[x].type) for x in sql_fields.keys() if x in table
        )
        # make sure all field names are lower case to avoid
        # migrations because of case change
        sql_fields = dict(map(self._fix, iteritems(sql_fields)))
        sql_fields_old = dict(map(self._fix, iteritems(sql_fields_old)))
        sql_fields_aux = dict(map(self._fix, iteritems(sql_fields_aux)))

        table_rname = table._rname
        if self.dbengine == "oracle":
            table_rname = self.dialect.quote(table_rname)

        if db._debug:
            db.logger.debug("migrating %s to %s" % (sql_fields_old, sql_fields))

        keys = list(sql_fields.keys())
        for key in sql_fields_old:
            if key not in keys:
                keys.append(key)
        new_add = self.dialect.concat_add(table_rname)

        metadata_change = False
        sql_fields_current = copy.copy(sql_fields_old)
        for key in keys:
            query = None
            if key not in sql_fields_old:
                sql_fields_current[key] = sql_fields[key]
                if self.dbengine in ("postgres",) and sql_fields[key][
                    "type"
                ].startswith("geometry"):
                    # 'sql' == ftype in sql
                    query = [sql_fields[key]["sql"]]
                else:
                    query = [
                        "ALTER TABLE %s ADD %s %s;"
                        % (
                            table_rname,
                            sql_fields[key]["rname"],
                            sql_fields_aux[key]["sql"].replace(", ", new_add),
                        )
                    ]
                metadata_change = True
            elif self.dbengine in ("sqlite", "spatialite"):
                if key in sql_fields:
                    sql_fields_current[key] = sql_fields[key]
                    # Field rname has changes, add new column
                    if (
                        sql_fields[key]["raw_rname"].lower()
                        != sql_fields_old[key]["raw_rname"].lower()
                    ):
                        tt = sql_fields_aux[key]["sql"].replace(", ", new_add)
                        query = [
                            "ALTER TABLE %s ADD %s %s;"
                            % (table_rname, sql_fields[key]["rname"], tt),
                            "UPDATE %s SET %s=%s;"
                            % (
                                table_rname,
                                sql_fields[key]["rname"],
                                sql_fields_old[key]["rname"],
                            ),
                        ]
                metadata_change = True
            elif key not in sql_fields:
                del sql_fields_current[key]
                ftype = sql_fields_old[key]["type"]
                if self.dbengine == "postgres" and ftype.startswith("geometry"):
                    geotype, parms = ftype[:-1].split("(")
                    schema = parms.split(",")[0]
                    query = [
                        "SELECT DropGeometryColumn ('%(schema)s', \
                             '%(table)s', '%(field)s');"
                        % dict(
                            schema=schema,
                            table=table._raw_rname,
                            field=sql_fields_old[key]["raw_rname"],
                        )
                    ]
                else:
                    query = [drop_expr % (table._rname, sql_fields_old[key]["rname"])]
                metadata_change = True
            # The field has a new rname, temp field is not needed
            elif (
                sql_fields[key]["raw_rname"].lower()
                != sql_fields_old[key]["raw_rname"].lower()
            ):
                sql_fields_current[key] = sql_fields[key]
                tt = sql_fields_aux[key]["sql"].replace(", ", new_add)
                query = [
                    "ALTER TABLE %s ADD %s %s;"
                    % (table_rname, sql_fields[key]["rname"], tt),
                    "UPDATE %s SET %s=%s;"
                    % (
                        table_rname,
                        sql_fields[key]["rname"],
                        sql_fields_old[key]["rname"],
                    ),
                    drop_expr % (table_rname, sql_fields_old[key]["rname"]),
                ]
                metadata_change = True
            elif (
                sql_fields[key]["sql"] != sql_fields_old[key]["sql"]
                and not isinstance(field_types.get(key), SQLCustomType)
                and not sql_fields[key]["type"].startswith("reference")
                and not sql_fields[key]["type"].startswith("double")
                and not sql_fields[key]["type"].startswith("id")
            ):
                sql_fields_current[key] = sql_fields[key]
                tt = sql_fields_aux[key]["sql"].replace(", ", new_add)
                key_tmp = self.dialect.quote(key + "__tmp")
                query = [
                    "ALTER TABLE %s ADD %s %s;" % (table_rname, key_tmp, tt),
                    "UPDATE %s SET %s=%s;"
                    % (table_rname, key_tmp, sql_fields_old[key]["rname"]),
                    drop_expr % (table_rname, sql_fields_old[key]["rname"]),
                    "ALTER TABLE %s ADD %s %s;"
                    % (table_rname, sql_fields[key]["rname"], tt),
                    "UPDATE %s SET %s=%s;"
                    % (table_rname, sql_fields[key]["rname"], key_tmp),
                    drop_expr % (table_rname, key_tmp),
                ]
                metadata_change = True
            elif sql_fields[key] != sql_fields_old[key]:
                sql_fields_current[key] = sql_fields[key]
                metadata_change = True

            if query:
                self.log(
                    "timestamp: %s\n" % datetime.datetime.today().isoformat(), table
                )
                for sub_query in query:
                    self.log(sub_query + "\n", table)
                    if fake_migrate:
                        if db._adapter.commit_on_alter_table:
                            self.save_dbt(table, sql_fields_current)
                        self.log("faked!\n", table)
                    else:
                        self.adapter.execute(sub_query)
                        # Caveat: mysql, oracle and firebird
                        # do not allow multiple alter table
                        # in one transaction so we must commit
                        # partial transactions and
                        # update table._dbt after alter table.
                        if db._adapter.commit_on_alter_table:
                            db.commit()
                            self.save_dbt(table, sql_fields_current)
                            self.log("success!\n", table)

            elif metadata_change:
                self.save_dbt(table, sql_fields_current)

        if metadata_change and not (query and db._adapter.commit_on_alter_table):
            db.commit()
            self.save_dbt(table, sql_fields_current)
            self.log("success!\n", table)

    def save_dbt(self, table, sql_fields_current):
        tfile = self.file_open(table._dbt, "wb")
        pickle.dump(sql_fields_current, tfile)
        self.file_close(tfile)

    def log(self, message, table=None):
        isabs = None
        logfilename = self.adapter.adapter_args.get("logfile", "sql.log")
        writelog = bool(logfilename)
        if writelog:
            isabs = os.path.isabs(logfilename)
        if table and table._dbt and writelog and self.adapter.folder:
            if isabs:
                table._loggername = logfilename
            else:
                table._loggername = pjoin(self.adapter.folder, logfilename)
            logfile = self.file_open(table._loggername, "ab")
            logfile.write(to_bytes(message))
            self.file_close(logfile)

    @staticmethod
    def file_open(filename, mode="rb", lock=True):
        # to be used ONLY for files that on GAE may not be on filesystem
        if lock:
            fileobj = portalocker.LockedFile(filename, mode)
        else:
            fileobj = open(filename, mode)
        return fileobj

    @staticmethod
    def file_close(fileobj):
        # to be used ONLY for files that on GAE may not be on filesystem
        if fileobj:
            fileobj.close()

    @staticmethod
    def file_delete(filename):
        os.unlink(filename)

    @staticmethod
    def file_exists(filename):
        # to be used ONLY for files that on GAE may not be on filesystem
        return exists(filename)


class InDBMigrator(Migrator):
    def file_exists(self, filename):
        return DatabaseStoredFile.exists(self.db, filename)

    def file_open(self, filename, mode="rb", lock=True):
        return DatabaseStoredFile(self.db, filename, mode)

    @staticmethod
    def file_close(fileobj):
        fileobj.close_connection()

    def file_delete(self, filename):
        query = "DELETE FROM web2py_filesystem WHERE path='%s'" % filename
        self.db.executesql(query)
        self.db.commit()



import re

from .._compat import basestring, integer_types
from ..adapters.oracle import Oracle
from . import dialects, sqltype_for
from .base import SQLDialect


@dialects.register_for(Oracle)
class OracleDialect(SQLDialect):
    false_exp = "1=0"
    true_exp = "1=1"

    @sqltype_for("string")
    def type_string(self):
        return "VARCHAR2(%(length)s)"

    @sqltype_for("text")
    def type_text(self):
        return "CLOB"

    @sqltype_for("integer")
    def type_integer(self):
        return "INT"

    @sqltype_for("bigint")
    def type_bigint(self):
        return "NUMBER"

    @sqltype_for("double")
    def type_double(self):
        return "BINARY_DOUBLE"

    @sqltype_for("time")
    def type_time(self):
        return "TIME(8)"

    @sqltype_for("datetime")
    def type_datetime(self):
        return "DATE"

    @sqltype_for("id")
    def type_id(self):
        return "NUMBER PRIMARY KEY"

    @sqltype_for("reference")
    def type_reference(self):
        return (
            "NUMBER, CONSTRAINT %(constraint_name)s FOREIGN KEY "
            + "(%(field_name)s) REFERENCES %(foreign_key)s ON DELETE "
            + "%(on_delete_action)s"
        )

    @sqltype_for("reference FK")
    def type_reference_fk(self):
        return (
            ", CONSTRAINT FK_%(constraint_name)s FOREIGN KEY "
            + "(%(field_name)s) REFERENCES %(foreign_key)s "
            + "ON DELETE %(on_delete_action)s"
        )

    @sqltype_for("reference TFK")
    def type_reference_tfk(self):
        return (
            " CONSTRAINT FK_%(constraint_name)s_PK FOREIGN KEY "
            + "(%(field_name)s) REFERENCES %(foreign_table)s"
            + "(%(foreign_key)s) ON DELETE %(on_delete_action)s"
        )

    def left_join(self, val, query_env={}):
        if not isinstance(val, basestring):
            val = self.expand(val, query_env=query_env)
        return "LEFT OUTER JOIN %s" % val

    @property
    def random(self):
        return "dbms_random.value"

    def cast(self, first, second, query_env={}):
        if second == "CLOB":
            return "TO_CHAR(%s)" % self.expand(first, query_env=query_env)
        return "CAST(%s)" % self._as(first, second, query_env)

    def mod(self, first, second, query_env={}):
        return "MOD(%s,%s)" % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env),
        )

    def extract(self, first, what, query_env={}):
        if what == "hour":
            return "TO_CHAR(%s, 'HH24')" % self.expand(first, query_env=query_env)
        if what == "minute":
            return "TO_CHAR(%s, 'MI')" % self.expand(first, query_env=query_env)
        if what == "second":
            return "TO_CHAR(%s, 'SS')" % self.expand(first, query_env=query_env)
        return "EXTRACT(%s FROM %s)" % (what, self.expand(first, query_env=query_env))

    def epoch(self, val, query_env={}):
        return "(%s - DATE '1970-01-01')*24*60*60" % self.expand(
            val, query_env=query_env
        )

    def quote(self, val):
        if not (val[0] == '"' and val[-1] == '"'):
            return self.quote_template % val
        return val

    def _as(self, first, second, query_env={}):
        return "%s %s" % (self.expand(first, query_env), self.quote(second))

    def alias(self, original, new):
        return "%s %s" % (original, self.quote(new))

    def writing_alias(self, table):
        return self.sql_fullref(table)

    def sqlsafe(self, field):
        if field._table is None:
            raise SyntaxError("Field %s is not bound to any table" % field.name)
        return self.quote(field._table.sql_shortref) + "." + self.quote(field._rname)

    def longname(self, field):
        if field._table is None:
            raise SyntaxError("Field %s is not bound to any table" % field.name)
        return self.quote(field._table._tablename) + "." + self.quote(field.name)

    def sql_fullref(self, table):
        if table._tablename == table._dalname:
            return self.quote(table._rname)
        return self.adapter.sqlsafe_table(table._tablename, table._rname)

    def trigger_name(self, tablename):
        return "%s_trigger" % tablename

    def sequence_name(self, tablename):
        if tablename[0] == '"':
            # manually written quotes, typically in case-sensitive rname
            tablename = tablename[1:-1]
        # truncate to max length
        return self.quote(("%s_sequence" % tablename)[0:29])

    def constraint_name(self, table, fieldname):
        if table[0] == '"':
            # manually written quotes, typically in case-sensitive rname
            table = table[1:-1]
        constraint_name = super(OracleDialect, self).constraint_name(table, fieldname)
        if len(constraint_name) > 30:
            constraint_name = "%s_%s__constraint" % (table[:10], fieldname[:7])
        return constraint_name

    def primary_key(self, key):
        if len(re.split(r",\s*", key)) > 1:
            return "PRIMARY KEY(%s)" % ", ".join(
                [self.quote(k) for k in re.split(r",\s*", key)]
            )
        return "PRIMARY KEY(%s)" % key

    def not_null(self, default, field_type):
        return "DEFAULT %s NOT NULL" % self.adapter.represent(default, field_type)

    def not_null(self, default, field_type):
        return "NOT NULL DEFAULT %s" % self.adapter.represent(default, field_type)

    def eq(self, first, second=None, query_env={}):
        if (first.type == "text" or first.type[:4] == "list") and second:
            return "(TO_CHAR(%s) = %s)" % (
                self.expand(first, query_env=query_env),
                self.expand(second, first.type, query_env=query_env),
            )
        return super(OracleDialect, self).eq(first, second, query_env)

    def regexp(self, first, second, query_env={}):
        return "REGEXP_LIKE(%s, %s)" % (
            self.expand(first, query_env=query_env),
            self.expand(second, "string", query_env=query_env),
        )

    def insert(self, table, fields, values):
        return "INSERT INTO %s(%s) VALUES (%s);" % (self.quote(table), fields, values)

    def insert_empty(self, table):
        return "INSERT INTO %s VALUES (DEFAULT);" % table

    def _select_aux(self, sql, fields, attributes, colnames):
        return super._select_aux(sql, fields, attributes, colnames)

    def select(
        self,
        fields,
        tables,
        where=None,
        groupby=None,
        having=None,
        orderby=None,
        limitby=None,
        distinct=False,
        for_update=False,
        with_cte=None,
    ):
        dst, whr, grp, order, limit, offset, upd = "", "", "", "", "", "", ""
        if distinct is True:
            dst = " DISTINCT"
        elif distinct:
            dst = " DISTINCT ON (%s)" % distinct
        if where:
            whr = " %s" % self.where(where)
        if groupby:
            grp = " GROUP BY %s" % groupby
            if having:
                grp += " HAVING %s" % having
        if orderby:
            order = " ORDER BY %s" % orderby

        if with_cte:
            recursive, cte = with_cte
            recursive = " RECURSIVE" if recursive else ""
            with_cte = "WITH%s %s " % (recursive, cte)
        else:
            with_cte = ""

        if limitby:
            (lmin, lmax) = limitby
            if whr:
                whr2 = whr + " AND w_row > %i" % lmin
            else:
                whr2 = self.where("w_row > %i" % lmin)
            return """
                %sSELECT%s * FROM (
                    SELECT w_tmp.*, ROWNUM w_row FROM (
                        SELECT %s FROM %s%s%s%s
                    ) w_tmp
                ) WHERE w_row<=%i and w_row>%i
            """ % (
                with_cte,
                dst,
                fields,
                tables,
                whr,
                grp,
                order,
                lmax,
                lmin,
            )
        if for_update:
            upd = " FOR UPDATE"
        return "%sSELECT%s %s FROM %s%s%s%s%s%s%s;" % (
            with_cte,
            dst,
            fields,
            tables,
            whr,
            grp,
            order,
            limit,
            offset,
            upd,
        )

    def drop_table(self, table, mode):
        sequence_name = table._sequence_name
        if mode and mode.upper() == "CASCADE":
            mode = "CASCADE CONSTRAINTS"
        drops = [
            "DROP TABLE %s %s;" % (self.quote(table._rname), mode),
        ]
        if "_id" in table:
            drops.append("DROP SEQUENCE %s;" % sequence_name)
        return drops



from .._compat import integer_types, to_unicode
from ..adapters.google import GoogleDatastore
from ..helpers.serializers import serializers
from . import for_type, pre, representers
from .base import NoSQLRepresenter

long = integer_types[-1]


@representers.register_for(GoogleDatastore)
class GoogleDatastoreRepresenter(NoSQLRepresenter):
    @pre(is_breaking=True)
    def _keep_lists_for_in_operator(self, obj, field_type):
        if (
            isinstance(field_type, str)
            and isinstance(obj, list)
            and not field_type.startswith("list:")
        ):
            return True, [self.adapter.represent(v, field_type) for v in obj]
        return False, obj

    @for_type("json")
    def _json(self, value):
        return serializers.json(value)

    @for_type("list:integer")
    def _list_integer(self, value):
        if not isinstance(value, list):
            return int(value)
        values = self._represent_list(value)
        return list(map(int, values))

    @for_type("list:string")
    def _list_string(self, value):
        if not isinstance(value, list):
            return str(value)
        values = self._represent_list(value)
        return list(map(to_unicode, values))

    @for_type("list:reference")
    def _list_reference(self, value):
        if not isinstance(value, list):
            return long(value)
        values = self._represent_list(value)
        return list(map(long, values))



import re

from ..utils import split_uri_args
from . import adapters, with_connection
from .base import SQLAdapter


@adapters.register_for("mysql:mysqlconnector")
@adapters.register_for("mysql:pymysql")
@adapters.register_for("mysql:MySQLdb")
@adapters.register_for("mysql")
class MySQL(SQLAdapter):
    dbengine = "mysql"
    drivers = ("MySQLdb", "pymysql", "mysqlconnector")
    commit_on_alter_table = True
    support_distributed_transaction = True

    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]*|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)"
        r"(\?(?P<uriargs>.*))?$"
    )  # set_encoding and unix_socket

    def _initialize_(self):
        super(MySQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        host = m.group("host")
        uriargs = m.group("uriargs")
        if uriargs:
            uri_args = split_uri_args(uriargs, need_equal=True)
            charset = uri_args.get("set_encoding") or "utf8"
            socket = uri_args.get("unix_socket")
        else:
            charset = "utf8"
            socket = None
        # NOTE:
        # MySQLdb (see http://mysql-python.sourceforge.net/MySQLdb.html)
        # use UNIX sockets and named pipes by default if no host is given
        # or host is 'localhost'; as opposed to
        # pymysql (see https://pymysql.readthedocs.io/en/latest/modules/connections.html)
        # or mysqlconnector (see https://dev.mysql.com/doc/connectors/en/connector-python-connectargs.html)
        # driver, where you have to specify the socket explicitly.
        if not host and not socket:
            raise SyntaxError("Host or UNIX socket name required")
        db = m.group("db")
        port = int(m.group("port") or "3306")
        self.driver_args.update(user=user, db=db, charset=charset)
        if password is not None:
            self.driver_args["passwd"] = password
        if socket:
            self.driver_args["unix_socket"] = socket
        else:
            self.driver_args.update(host=host, port=port)

    def connector(self):
        cursor_buffered = self.driver_args.get("cursor_buffered")
        if cursor_buffered:
            del self.driver_args["cursor_buffered"]
        conn = self.driver.connect(**self.driver_args)
        if cursor_buffered:
            conn.cursor = lambda conn=conn: conn.cursor(buffered=True)
        return conn

    def after_connection(self):
        self.execute("SET FOREIGN_KEY_CHECKS=1;")
        self.execute("SET sql_mode='NO_BACKSLASH_ESCAPES';")

    def distributed_transaction_begin(self, key):
        self.execute("XA START;")

    @with_connection
    def prepare(self, key):
        self.execute("XA END;")
        self.execute("XA PREPARE;")

    @with_connection
    def commit_prepared(self, key):
        self.execute("XA COMMIT;")

    @with_connection
    def rollback_prepared(self, key):
        self.execute("XA ROLLBACK;")


@adapters.register_for("cubrid")
class Cubrid(MySQL):
    dbengine = "cubrid"
    drivers = ("cubriddb",)

    def _initialize_(self):
        super(Cubrid, self)._initialize_()
        del self.driver_args["charset"]



from .._gae import ndb
from ..adapters.google import GoogleDatastore
from ..helpers.gae import NDBDecimalProperty
from . import dialects, sqltype_for
from .base import NoSQLDialect


@dialects.register_for(GoogleDatastore)
class GoogleDatastoreDialect(NoSQLDialect):
    FILTER_OPTIONS = {
        "=": lambda a, b: a == b,
        ">": lambda a, b: a > b,
        "<": lambda a, b: a < b,
        "<=": lambda a, b: a <= b,
        ">=": lambda a, b: a >= b,
        "!=": lambda a, b: a != b,
        "in": lambda a, b: a.IN(b),
    }

    @sqltype_for("string")
    def type_string(self):
        return lambda **kwargs: ndb.StringProperty(**kwargs)

    @sqltype_for("boolean")
    def type_boolean(self):
        return ndb.BooleanProperty

    @sqltype_for("text")
    def type_text(self):
        return ndb.TextProperty

    @sqltype_for("json")
    def type_json(self):
        return self.types["text"]

    @sqltype_for("password")
    def type_password(self):
        return ndb.StringProperty

    @sqltype_for("blob")
    def type_blob(self):
        return ndb.BlobProperty

    @sqltype_for("upload")
    def type_upload(self):
        return self.types["password"]

    @sqltype_for("integer")
    def type_integer(self):
        return ndb.IntegerProperty

    @sqltype_for("bigint")
    def type_bigint(self):
        return self.types["integer"]

    @sqltype_for("float")
    def type_float(self):
        return ndb.FloatProperty

    @sqltype_for("double")
    def type_double(self):
        return self.types["float"]

    @sqltype_for("decimal")
    def type_decimal(self):
        return NDBDecimalProperty

    @sqltype_for("date")
    def type_date(self):
        return ndb.DateProperty

    @sqltype_for("time")
    def type_time(self):
        return ndb.TimeProperty

    @sqltype_for("datetime")
    def type_datetime(self):
        return ndb.DateTimeProperty

    @sqltype_for("id")
    def type_id(self):
        return None

    @sqltype_for("reference")
    def type_reference(self):
        return ndb.IntegerProperty

    @sqltype_for("list:integer")
    def type_list_integer(self):
        return lambda **kwargs: ndb.IntegerProperty(
            repeated=True, default=None, **kwargs
        )

    @sqltype_for("list:string")
    def type_list_string(self):
        return lambda **kwargs: ndb.StringProperty(
            repeated=True, default=None, **kwargs
        )

    @sqltype_for("list:reference")
    def type_list_reference(self):
        return lambda **kwargs: ndb.IntegerProperty(
            repeated=True, default=None, **kwargs
        )

    def _and(self, first, second, query_env={}):
        first = self.expand(first, query_env=query_env)
        second = self.expand(second, query_env=query_env)
        # none means lack of query (true)
        if first == None:
            return second
        return ndb.AND(first, second)

    def _or(self, first, second, query_env={}):
        first = self.expand(first, query_env=query_env)
        second = self.expand(second, query_env=query_env)
        # none means lack of query (true)
        if first == None or second == None:
            return None
        return ndb.OR(first, second)

    def __gaef(self, first, op, second):
        name = first.name if first.name != "id" else "key"
        if name == "key" and op in (">", "!=") and second in (0, "0", None):
            return None
        field = getattr(first.table._tableobj, name)
        value = self.adapter.represent(second, first.type, first._tablename)
        return self.FILTER_OPTIONS[op](field, value)

    def eq(self, first, second=None, query_env={}):
        return self.__gaef(first, "=", second)

    def ne(self, first, second=None, query_env={}):
        return self.__gaef(first, "!=", second)

    def lt(self, first, second=None, query_env={}):
        return self.__gaef(first, "<", second)

    def lte(self, first, second=None, query_env={}):
        return self.__gaef(first, "<=", second)

    def gt(self, first, second=None, query_env={}):
        return self.__gaef(first, ">", second)

    def gte(self, first, second=None, query_env={}):
        return self.__gaef(first, ">=", second)

    def invert(self, first, query_env={}):
        return "-%s" % first.name

    def comma(self, first, second, query_env={}):
        return "%s,%s" % (first, second)

    def belongs(self, first, second, query_env={}):
        if not isinstance(second, (list, tuple, set)):
            raise SyntaxError("Not supported")
        if not isinstance(second, list):
            second = list(second)
        if len(second) == 0:
            # return a filter which will return a null set
            f = self.eq(first, 0)
            f.filter_all = True
            return f
        return self.__gaef(first, "in", second)

    def contains(self, first, second, case_sensitive=True, query_env={}):
        # silently ignoring: GAE can only do case sensitive matches!
        if not first.type.startswith("list:"):
            raise SyntaxError("Not supported")
        return self.__gaef(first, "=", second)

    def _not(self, val, query_env={}):
        op, f, s = val.op, val.first, val.second
        if op in [self._or, self._and]:
            not_op = self._and if op == self._or else self._or
            rv = not_op(self._not(f), self._not(s))
        elif op == self.eq:
            rv = self.__gaef(f, "!=", s)
        elif op == self.ne:
            rv = self.__gaef(f, "=", s)
        elif op == self.lt:
            rv = self.__gaef(f, ">=", s)
        elif op == self.lte:
            rv = self.__gaef(f, ">", s)
        elif op == self.gt:
            rv = self.__gaef(f, "<=", s)
        elif op == self.gte:
            rv = self.__gaef(f, "<", s)
        else:
            # TODO the IN operator must be split into a sequence of
            # (field!=value) AND (field!=value) AND ...
            raise NotImplementedError
        return rv



# -*- coding: utf-8 -*-

"""
| This file is part of the web2py Web Framework
| Copyrighted by Massimo Di Pierro <mdipierro@cs.depaul.edu>
| License: BSD

Takes care of adapting pyDAL to web2py's needs
-----------------------------------------------
"""

from . import validators


def default_validators(db, field):
    """
    Field type validation, using web2py's validators mechanism.

    makes sure the content of a field is in line with the declared
    fieldtype
    """

    field_type = field.type
    field_unique = field.unique
    field_notnull = field.notnull

    is_ref = field_type.startswith("reference")
    is_list = field_type.startswith("list:")
    if is_ref or field_type.startswith("list:reference"):
        table_field = field_type[10 if is_ref else 15 :].split(".", 1)
        table_name = table_field[0]
        field_name = table_field[-1]
        requires = None
        if table_name in db.tables:
            referenced = db[table_name]
            if len(table_field) == 1:
                requires = validators.IS_IN_DB(
                    db,
                    referenced._id,
                    label=getattr(referenced, "_format", None),
                    multiple=is_list,
                )
            elif field_name in referenced.fields:
                requires = validators.IS_IN_DB(
                    db,
                    getattr(referenced, field_name),
                    label=getattr(referenced, "_format", None),
                    multiple=is_list,
                )
        if requires:
            if field_unique:
                requires._and = validators.IS_NOT_IN_DB(db, field)
            if not field_notnull:
                requires = validators.IS_EMPTY_OR(requires)
        return requires

    if isinstance(field.options, (list, tuple)):
        requires = validators.IS_IN_SET(field.options, multiple=is_list)
    else:
        requires = []
        if field_type in (("string", "text", "password")):
            requires.append(validators.IS_LENGTH(field.length))
        elif field_type == "json":
            requires.append(validators.IS_EMPTY_OR(validators.IS_JSON()))
        elif field_type == "double" or field_type == "float":
            requires.append(validators.IS_FLOAT_IN_RANGE(-1e100, 1e100))
        elif field_type == "integer":
            requires.append(validators.IS_INT_IN_RANGE(-(2 ** 31), 2 ** 31))
        elif field_type == "bigint":
            requires.append(validators.IS_INT_IN_RANGE(-(2 ** 63), 2 ** 63))
        elif field_type.startswith("decimal"):
            requires.append(validators.IS_DECIMAL_IN_RANGE(-(10 ** 10), 10 ** 10))
        elif field_type == "date":
            requires.append(validators.IS_DATE())
        elif field_type == "time":
            requires.append(validators.IS_TIME())
        elif field_type == "datetime":
            requires.append(validators.IS_DATETIME())

        if field_unique:
            requires.insert(0, validators.IS_NOT_IN_DB(db, field))
        if (field_notnull or field_unique) and field_type not in (
            "boolean",
            "password",
            "string",
            "text",
            "upload",
        ):
            requires.insert(0, validators.IS_NOT_EMPTY())
        elif not field_notnull and not field_unique and requires:
            null = "" if field.type == "password" else None
            requires[0] = validators.IS_EMPTY_OR(requires[0], null=null)

        if len(requires) == 1:
            requires = requires[0]

    return requires or None



from ..adapters.google import GoogleDatastore
from . import for_type, parsers
from .base import BasicParser, JSONParser


@parsers.register_for(GoogleDatastore)
class GoogleDatastoreParser(BasicParser, JSONParser):
    @for_type("id")
    def _id(self, value):
        return value

    @for_type("boolean")
    def _boolean(self, value):
        return value



from ._compat import unittest
from ._adapt import DEFAULT_URI, IS_GAE, IS_IMAP, drop
from pydal._compat import integer_types
from pydal import DAL, Field
from pydal.helpers.methods import smart_query


@unittest.skipIf(IS_IMAP, "Skip nosql")
class TestSmartQuery(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])

        # -----------------------------------------------------------------------------
        # Seems further imports are required for the commented field types below

        # db.define_table('referred_table',
        #                 Field('represent_field', 'string'))
        # NOTE : Don't forget to uncomment the line # drop(db.referred_table) at the very end below
        #        if the above are uncommented

        db.define_table(
            "a_table",
            Field("string_field", "string"),
            Field("text_field", "text"),
            Field("boolean_field", "boolean"),
            Field("integer_field", "integer"),
            Field("double_field", "double"),
            # Field('decimal_field', 'decimal'),
            # Field('date_field', 'date'),
            # Field('time_field', 'time'),
            # Field('datetime_field', 'datetime'),
            # Field('reference_field', 'reference referred_table'),
            # Field('list_string_field', 'list:string'),
            # Field('list_integer_field', 'list:integer'),
            # Field('list_reference_field', 'list:reference referred_table')
        )

        fields = [
            db.a_table.id,
            db.a_table.string_field,
            db.a_table.text_field,
            db.a_table.boolean_field,
            db.a_table.integer_field,
            db.a_table.double_field,
            # db.a_table.decimal_field,
            # db.a_table.date_field,
            # db.a_table.time_field,
            # db.a_table.reference_field,
            # db.a_table.list_string_field,
            # db.a_table.list_integer_field,
            # db.a_table.list_reference_field
        ]
        # -----------------------------------------------------------------------------

        # -----------------------------------------------------------------------------
        # Test with boolean field
        # Operator under test
        # operators = \
        #         [(' starts with ','startswith'),
        #          (' ends with ','endswith'),
        #          ('contains', 'N/A'),
        #          ('like', 'N/A')
        #          ]
        #
        #

        keywords = "a_table.boolean_field = True"
        q = db.a_table.boolean_field == True
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        if not IS_GAE:
            # Test string field query
            # starts with
            keywords = 'a_table.string_field starts with "pydal"'
            q = db.a_table.string_field.startswith("pydal")
            smart_q = smart_query(fields, keywords)
            self.assertEqual(smart_q, q)

            # ends with
            keywords = 'a_table.string_field ends with "Rocks!!"'
            q = db.a_table.string_field.endswith("Rocks!!")
            smart_q = smart_query(fields, keywords)
            self.assertEqual(smart_q, q)

            # contains
            keywords = 'a_table.string_field contains "Rocks"'
            q = db.a_table.string_field.contains("Rocks")
            smart_q = smart_query(fields, keywords)
            self.assertEqual(smart_q, q)

        # Don't work for some reason
        # # like
        # keywords = 'a_table.string_field like "%Rocks%"'
        # q = (db.a_table.string_field.like('%Rocks%'))
        # smart_q = smart_query(fields, keywords)
        # self.assertTrue(smart_q == q)
        # -----------------------------------------------------------------------------

        # -----------------------------------------------------------------------------
        # Tests with integer field
        # For generating these tests
        # def generate_tests():
        #     operators = \
        #         [('=', '='),
        #          ('==', '='),
        #          (' is ','='),
        #          (' equal ', '='),
        #          (' equals ', '='),
        #          (' equal to ', '='),
        #          ('<>', '!='),
        #          (' not equal ', '!='),
        #          (' not equal to ', '!='),
        #          ('<', '<'),
        #          (' less than ', '<'),
        #          ('<=', '<='),
        #          ('=<', '<='),
        #          (' less or equal ', '<='),
        #          (' less or equal than ', '<='),
        #          (' equal or less ', '<='),
        #          (' equal or less than ', '<='),
        #          ('>', '>'),
        #          (' greater than ', '>'),
        #          ('=>', '>='),
        #          ('>=', '>='),
        #          (' greater or equal ', '>='),
        #          (' greater or equal than ', '>='),
        #          (' equal or greater ', '>='),
        #          (' equal or greater than ', '>=')]  # JUST APPEND MORE OPERATORS HERE
        #
        #     for op in operators:
        #         print """
        #         # {op}
        #         keywords = 'a_table.integer_field {test_op} 1'
        #         q = (db.a_table.integer_field {result_op} 1)
        #         smart_q = smart_query(fields, keywords)
        #         self.assertTrue(smart_q == q)""".format(op=op,
        #                                                 test_op=op[0],
        #                                                 result_op='==' if op[1] == '=' else op[1])

        # ('=', '=')
        keywords = "a_table.integer_field = 1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # ('==', '=')
        keywords = "a_table.integer_field == 1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' is ','=')
        keywords = "a_table.integer_field is 1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal ', '=')
        keywords = "a_table.integer_field  equal  1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equals ', '=')
        keywords = "a_table.integer_field  equals  1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal to ', '=')
        keywords = "a_table.integer_field  equal to  1"
        q = db.a_table.integer_field == 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # This one not allow over integer it seems
        # # ('<>', '!=')
        # keywords = 'a_table.integer_field <> 1'
        # q = (db.a_table.integer_field != 1)
        # smart_q = smart_query(fields, keywords)
        # self.assertTrue(smart_q == q)

        # (' not equal ', '!=')
        keywords = "a_table.integer_field  not equal  1"
        q = db.a_table.integer_field != 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' not equal to ', '!=')
        keywords = "a_table.integer_field  not equal to  1"
        q = db.a_table.integer_field != 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # ('<', '<')
        keywords = "a_table.integer_field < 1"
        q = db.a_table.integer_field < 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' less than ', '<')
        keywords = "a_table.integer_field  less than  1"
        q = db.a_table.integer_field < 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # ('<=', '<=')
        keywords = "a_table.integer_field <= 1"
        q = db.a_table.integer_field <= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # This one is invalid, maybe we should remove it from smart_query
        # # ('=<', '<=')
        # keywords = 'a_table.integer_field =< 1'
        # q = (db.a_table.integer_field <= 1)
        # smart_q = smart_query(fields, keywords)
        # self.assertTrue(smart_q == q)

        # (' less or equal ', '<=')
        keywords = "a_table.integer_field  less or equal  1"
        q = db.a_table.integer_field <= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' less or equal than ', '<=')
        keywords = "a_table.integer_field  less or equal than  1"
        q = db.a_table.integer_field <= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal or less ', '<=')
        keywords = "a_table.integer_field  equal or less  1"
        q = db.a_table.integer_field <= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal or less than ', '<=')
        keywords = "a_table.integer_field  equal or less than  1"
        q = db.a_table.integer_field <= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # ('>', '>')
        keywords = "a_table.integer_field > 1"
        q = db.a_table.integer_field > 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' greater than ', '>')
        keywords = "a_table.integer_field  greater than  1"
        q = db.a_table.integer_field > 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # This one is invalid, maybe we should remove it from smart_query
        # # ('=>', '>=')
        # keywords = 'a_table.integer_field => 1'
        # q = (db.a_table.integer_field >= 1)
        # smart_q = smart_query(fields, keywords)
        # self.assertTrue(smart_q == q)

        # ('>=', '>=')
        keywords = "a_table.integer_field >= 1"
        q = db.a_table.integer_field >= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' greater or equal ', '>=')
        keywords = "a_table.integer_field  greater or equal  1"
        q = db.a_table.integer_field >= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' greater or equal than ', '>=')
        keywords = "a_table.integer_field  greater or equal than  1"
        q = db.a_table.integer_field >= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal or greater ', '>=')
        keywords = "a_table.integer_field  equal or greater  1"
        q = db.a_table.integer_field >= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)

        # (' equal or greater than ', '>=')
        keywords = "a_table.integer_field  equal or greater than  1"
        q = db.a_table.integer_field >= 1
        smart_q = smart_query(fields, keywords)
        self.assertEqual(smart_q, q)
        # -----------------------------------------------------------------------------

        # -----------------------------------------------------------------------------
        # Belongs and not belongs

        # NOTE : The below tests don't works
        # Issue : https://github.com/web2py/pydal/issues/161

        # (' in ', 'belongs') -> field.belongs(1, 2, 3)
        # keywords = 'a_table.integer_field in "1, 2, 3"'
        # q = (db.a_table.integer_field.belongs([1, 2, 3]))
        # smart_q = smart_query(fields, keywords)
        # self.assertEqual(smart_q, q)

        # keywords = 'a_table.id in "1, 2, 3"'
        # q = (db.a_table.id.belongs([1, 2, 3]))
        # smart_q = smart_query(fields, keywords)
        # self.assertEqual(smart_q, q)
        #
        # # (' not in ' , 'notbelongs'),
        # keywords = 'a_table.integer_field not in "1, 2, 3"'
        # q = (~db.a_table.id.belongs([1, 2, 3]))
        # smart_q = smart_query(fields, keywords)
        # self.assertTrue(smart_q == q)

        # -----------------------------------------------------------------------------
        # cleanup table
        drop(db.a_table)
        # drop(db.referred_table)
        # -----------------------------------------------------------------------------


if __name__ == "__main__":
    unittest.main()



import copy
import random
from datetime import datetime

from .._compat import basestring, long
from ..exceptions import NotOnNOSQLError
from ..helpers.classes import SQLALL, FakeCursor, Reference
from ..helpers.methods import use_common_filters, xorify
from ..objects import Expression, Field, Query, Row
from . import adapters
from .base import NoSQLAdapter

try:
    from bson import Binary
    from bson.binary import USER_DEFINED_SUBTYPE
except:

    class Binary(object):
        pass

    USER_DEFINED_SUBTYPE = 0


@adapters.register_for("rethinkdb")
class RethinkDB(NoSQLAdapter):
    dbengine = "rethinkdb"
    drivers = ("rethinkdb",)

    def find_driver(self):
        super(RethinkDB, self).find_driver()
        if "fake_version" in self.driver_args:
            version = self.driver_args["fake_version"]
        else:
            from rethinkdb.version import version
        if int(version.split(".")[0]) < 2:
            raise RuntimeError(
                "pydal requires rethink version >= 2.0, found '%s'" % version
            )

    def _initialize_(self):
        super(RethinkDB, self)._initialize_()
        #: uri parse
        from pymongo import uri_parser

        m = uri_parser.parse_uri(self.uri)
        if isinstance(m, tuple):
            m = {"database": m[1]}
        if m.get("database") is None:
            raise SyntaxError("Database is required!")
        self._driver_db = m["database"]
        self.db_codec = "UTF-8"
        self.get_connection()

    #### XXXX

    def connector(self):
        conn = self.driver.MongoClient(self.uri, w=self.safe)[self._driver_db]
        conn.cursor = lambda: FakeCursor()
        conn.close = lambda: None
        conn.commit = lambda: None
        return conn

    def _after_first_connection(self):
        #: server version
        self._server_version = self.connection.command("serverStatus")["version"]
        self.server_version = tuple([int(x) for x in self._server_version.split(".")])
        self.server_version_major = (
            self.server_version[0] + self.server_version[1] / 10.0
        )

    def object_id(self, arg=None):
        """Convert input to a valid Mongodb ObjectId instance

        self.object_id("<random>") -> ObjectId (not unique) instance"""
        if not arg:
            arg = 0
        if isinstance(arg, basestring):
            # we assume an integer as default input
            rawhex = len(arg.replace("0x", "").replace("L", "")) == 24
            if arg.isdigit() and (not rawhex):
                arg = int(arg)
            elif arg == "<random>":
                arg = int(
                    "0x%s"
                    % "".join([random.choice("0123456789abcdef") for x in range(24)]),
                    0,
                )
            elif arg.isalnum():
                if not arg.startswith("0x"):
                    arg = "0x%s" % arg
                try:
                    arg = int(arg, 0)
                except ValueError as e:
                    raise ValueError("invalid objectid argument string: %s" % e)
            else:
                raise ValueError(
                    "Invalid objectid argument string. "
                    + "Requires an integer or base 16 value"
                )
        elif isinstance(arg, self.ObjectId):
            return arg
        elif isinstance(arg, (Row, Reference)):
            return self.object_id(long(arg["id"]))
        elif not isinstance(arg, (int, long)):
            raise TypeError(
                "object_id argument must be of type ObjectId or an objectid "
                + "representable integer (type %s)" % type(arg)
            )
        hexvalue = hex(arg)[2:].rstrip("L").zfill(24)
        return self.ObjectId(hexvalue)

    def _get_collection(self, tablename, safe=None):
        ctable = self.connection[tablename]
        if safe is not None and safe != self.safe:
            wc = self.WriteConcern(w=self._get_safe(safe))
            ctable = ctable.with_options(write_concern=wc)
        return ctable

    def _get_safe(self, val=None):
        if val is None:
            return self.safe
        return 1 if val else 0

    def _regex_select_as_parser(self, colname):
        return self.dialect.REGEX_SELECT_AS_PARSER.search(colname)

    @staticmethod
    def _parse_data(expression, attribute, value=None):
        if isinstance(expression, (list, tuple)):
            ret = False
            for e in expression:
                ret = Mongo._parse_data(e, attribute, value) or ret
            return ret
        if value is not None:
            try:
                expression._parse_data[attribute] = value
            except AttributeError:
                return None
        try:
            return expression._parse_data[attribute]
        except (AttributeError, TypeError):
            return None

    def _expand(self, expression, field_type=None, query_env={}):
        if isinstance(expression, Field):
            if expression.type == "id":
                result = "_id"
            else:
                result = expression.name
            if self._parse_data(expression, "pipeline"):
                # field names as part of expressions need to start with '$'
                result = "$" + result
        elif isinstance(expression, (Expression, Query)):
            first = expression.first
            second = expression.second
            if isinstance(first, Field) and "reference" in first.type:
                # cast to Mongo ObjectId
                if isinstance(second, (tuple, list, set)):
                    second = [self.object_id(item) for item in expression.second]
                else:
                    second = self.object_id(expression.second)
            op = expression.op
            optional_args = expression.optional_args or {}
            optional_args["query_env"] = query_env
            if second is not None:
                result = op(first, second, **optional_args)
            elif first is not None:
                result = op(first, **optional_args)
            elif isinstance(op, str):
                result = op
            else:
                result = op(**optional_args)
        elif isinstance(expression, Expansion):
            expression.query = self.expand(
                expression.query, field_type, query_env=query_env
            )
            result = expression
        elif isinstance(expression, (list, tuple)):
            result = [self.represent(item, field_type) for item in expression]
        elif field_type:
            result = self.represent(expression, field_type)
        else:
            result = expression
        return result

    def represent(self, obj, field_type):
        if isinstance(obj, self.ObjectId):
            return obj
        return super(Mongo, self).represent(obj, field_type)

    def truncate(self, table, mode, safe=None):
        ctable = self.connection[table._tablename]
        ctable.delete_many({})

    def count(self, query, distinct=None, snapshot=True):
        if not isinstance(query, Query):
            raise SyntaxError("Type '%s' not supported in count" % type(query))
        distinct_fields = []
        if distinct is True:
            distinct_fields = [x for x in query.first.table if x.name != "id"]
        elif distinct:
            if isinstance(distinct, Field):
                distinct_fields = [distinct]
            else:
                while isinstance(distinct, Expression) and isinstance(
                    distinct.second, Field
                ):
                    distinct_fields += [distinct.second]
                    distinct = distinct.first
                if isinstance(distinct, Field):
                    distinct_fields += [distinct]
            distinct = True
        expanded = Expansion(
            self, "count", query, fields=distinct_fields, distinct=distinct
        )
        ctable = expanded.get_collection()
        if not expanded.pipeline:
            return ctable.count(filter=expanded.query_dict)
        for record in ctable.aggregate(expanded.pipeline):
            return record["count"]
        return 0

    def select(self, query, fields, attributes, snapshot=False):
        attributes["snapshot"] = snapshot
        return self.__select(query, fields, **attributes)

    def __select(
        self,
        query,
        fields,
        left=False,
        join=False,
        distinct=False,
        orderby=False,
        groupby=False,
        having=False,
        limitby=False,
        orderby_on_limitby=True,
        for_update=False,
        outer_scoped=[],
        required=None,
        cache=None,
        cacheable=None,
        processor=None,
        snapshot=False,
    ):
        new_fields = []
        for item in fields:
            if isinstance(item, SQLALL):
                new_fields += item._table
            else:
                new_fields.append(item)
        fields = new_fields
        tablename = self.get_table(query, *fields)._tablename

        if for_update:
            self.db.logger.warning("Attribute 'for_update' unsuppored by MongoDB")
        if join or left:
            raise NotOnNOSQLError("Joins not supported on NoSQL databases")
        if required or cache or cacheable:
            self.db.logger.warning(
                "Attributes 'required', 'cache' and 'cacheable' are"
                + " unsuppored by MongoDB"
            )

        if limitby and orderby_on_limitby and not orderby:
            if groupby:
                orderby = groupby
            else:
                table = self.db[tablename]
                orderby = [
                    table[x]
                    for x in (
                        hasattr(table, "_primarykey") and table._primarykey or ["_id"]
                    )
                ]

        if not orderby:
            mongosort_list = []
        else:
            if snapshot:
                raise RuntimeError("snapshot and orderby are mutually exclusive")
            if isinstance(orderby, (list, tuple)):
                orderby = xorify(orderby)

            if str(orderby) == "<random>":
                # !!!! need to add 'random'
                mongosort_list = self.dialect.random
            else:
                mongosort_list = []
                for f in self.expand(orderby).split(","):
                    include = 1
                    if f.startswith("-"):
                        include = -1
                        f = f[1:]
                    if f.startswith("$"):
                        f = f[1:]
                    mongosort_list.append((f, include))

        expanded = Expansion(
            self,
            "select",
            query,
            fields or self.db[tablename],
            groupby=groupby,
            distinct=distinct,
            having=having,
        )
        ctable = self.connection[tablename]
        modifiers = {"snapshot": snapshot}

        if not expanded.pipeline:
            if limitby:
                limitby_skip, limitby_limit = limitby[0], int(limitby[1]) - 1
            else:
                limitby_skip = limitby_limit = 0
            mongo_list_dicts = ctable.find(
                expanded.query_dict,
                expanded.field_dicts,
                skip=limitby_skip,
                limit=limitby_limit,
                sort=mongosort_list,
                modifiers=modifiers,
            )
            null_rows = []
        else:
            if mongosort_list:
                sortby_dict = self.SON()
                for f in mongosort_list:
                    sortby_dict[f[0]] = f[1]
                expanded.pipeline.append({"$sort": sortby_dict})
            if limitby and limitby[1]:
                expanded.pipeline.append({"$limit": limitby[1]})
            if limitby and limitby[0]:
                expanded.pipeline.append({"$skip": limitby[0]})

            mongo_list_dicts = ctable.aggregate(expanded.pipeline)
            null_rows = [(None,)]

        rows = []
        # populate row in proper order
        # Here we replace ._id with .id to follow the standard naming
        colnames = []
        newnames = []
        for field in expanded.fields:
            if hasattr(field, "tablename"):
                if field.name in ("id", "_id"):
                    # Mongodb reserved uuid key
                    colname = (tablename + "." + "id", "_id")
                else:
                    colname = (field.longname, field.name)
            elif not isinstance(query, Expression):
                colname = (field.name, field.name)
            colnames.append(colname[1])
            newnames.append(colname[0])

        for record in mongo_list_dicts:
            row = []
            for colname in colnames:
                try:
                    value = record[colname]
                except:
                    value = None
                if self.server_version_major < 2.6:
                    # '$size' not present in server versions < 2.6
                    if isinstance(value, list) and "$addToSet" in colname:
                        value = len(value)

                row.append(value)
            rows.append(row)
        if not rows:
            rows = null_rows

        processor = processor or self.parse
        result = processor(rows, fields, newnames, blob_decode=True)
        return result

    def check_notnull(self, table, values):
        for fieldname in table._notnulls:
            if fieldname not in values or values[fieldname] is None:
                raise Exception("NOT NULL constraint failed: %s" % fieldname)

    def check_unique(self, table, values):
        if len(table._uniques) > 0:
            db = table._db
            unique_queries = []
            for fieldname in table._uniques:
                if fieldname in values:
                    value = values[fieldname]
                else:
                    value = table[fieldname].default
                unique_queries.append(
                    Query(db, self.dialect.eq, table[fieldname], value)
                )

            if len(unique_queries) > 0:
                unique_query = unique_queries[0]

                # if more than one field, build a query of ORs
                for query in unique_queries[1:]:
                    unique_query = Query(db, self.dialect._or, unique_query, query)

                if self.count(unique_query, distinct=False) != 0:
                    for query in unique_queries:
                        if self.count(query, distinct=False) != 0:
                            # one of the 'OR' queries failed, see which one
                            raise Exception(
                                "NOT UNIQUE constraint failed: %s" % query.first.name
                            )

    def insert(self, table, fields, safe=None):
        """Safe determines whether a asynchronous request is done or a
        synchronous action is done
        For safety, we use by default synchronous requests"""

        values = {}
        safe = self._get_safe(safe)
        ctable = self._get_collection(table._tablename, safe)

        for k, v in fields:
            if k.name not in ["id", "safe"]:
                fieldname = k.name
                fieldtype = table[k.name].type
                values[fieldname] = self.represent(v, fieldtype)

        # validate notnulls
        try:
            self.check_notnull(table, values)
        except Exception as e:
            if hasattr(table, "_on_insert_error"):
                return table._on_insert_error(table, fields, e)
            raise e

        # validate uniques
        try:
            self.check_unique(table, values)
        except Exception as e:
            if hasattr(table, "_on_insert_error"):
                return table._on_insert_error(table, fields, e)
            raise e

        # perform the insert
        result = ctable.insert_one(values)

        if result.acknowledged:
            Oid = result.inserted_id
            rid = Reference(long(str(Oid), 16))
            (rid._table, rid._record) = (table, None)
            return rid
        else:
            return None

    def update(self, table, query, fields, safe=None):
        # return amount of adjusted rows or zero, but no exceptions
        # @ related not finding the result
        if not isinstance(query, Query):
            raise RuntimeError("Not implemented")

        safe = self._get_safe(safe)
        if safe:
            amount = 0
        else:
            amount = self.count(query, distinct=False)
            if amount == 0:
                return amount

        expanded = Expansion(self, "update", query, fields)
        ctable = expanded.get_collection(safe)
        if expanded.pipeline:
            try:
                for doc in ctable.aggregate(expanded.pipeline):
                    result = ctable.replace_one({"_id": doc["_id"]}, doc)
                    if safe and result.acknowledged:
                        amount += result.matched_count
                return amount
            except Exception as e:
                # TODO Reverse update query to verify that the query succeeded
                raise RuntimeError("uncaught exception when updating rows: %s" % e)
        try:
            result = ctable.update_many(
                filter=expanded.query_dict, update={"$set": expanded.field_dicts}
            )
            if safe and result.acknowledged:
                amount = result.matched_count
            return amount
        except Exception as e:
            # TODO Reverse update query to verify that the query succeeded
            raise RuntimeError("uncaught exception when updating rows: %s" % e)

    def delete(self, table, query, safe=None):
        if not isinstance(query, Query):
            raise RuntimeError("query type %s is not supported" % type(query))

        safe = self._get_safe(safe)
        expanded = Expansion(self, "delete", query)
        ctable = expanded.get_collection(safe)
        if expanded.pipeline:
            deleted = [x["_id"] for x in ctable.aggregate(expanded.pipeline)]
        else:
            deleted = [x["_id"] for x in ctable.find(expanded.query_dict)]

        # find references to deleted items
        db = self.db
        cascade = []
        set_null = []
        for field in table._referenced_by:
            if field.type == "reference " + table._tablename:
                if field.ondelete == "CASCADE":
                    cascade.append(field)
                if field.ondelete == "SET NULL":
                    set_null.append(field)
        cascade_list = []
        set_null_list = []
        for field in table._referenced_by_list:
            if field.type == "list:reference " + table._tablename:
                if field.ondelete == "CASCADE":
                    cascade_list.append(field)
                if field.ondelete == "SET NULL":
                    set_null_list.append(field)

        # perform delete
        result = ctable.delete_many({"_id": {"$in": deleted}})
        if result.acknowledged:
            amount = result.deleted_count
        else:
            amount = len(deleted)

        # clean up any references
        if amount and deleted:
            # ::TODO:: test if deleted references cascade
            def remove_from_list(field, deleted, safe):
                for delete in deleted:
                    modify = {field.name: delete}
                    dtable = self._get_collection(field.tablename, safe)
                    dtable.update_many(filter=modify, update={"$pull": modify})

            # for cascaded items, if the reference is the only item in the
            # list, then remove the entire record, else delete reference
            # from the list
            for field in cascade_list:
                for delete in deleted:
                    modify = {field.name: [delete]}
                    dtable = self._get_collection(field.tablename, safe)
                    dtable.delete_many(filter=modify)
                remove_from_list(field, deleted, safe)
            for field in set_null_list:
                remove_from_list(field, deleted, safe)
            for field in cascade:
                db(field.belongs(deleted)).delete()
            for field in set_null:
                db(field.belongs(deleted)).update(**{field.name: None})
        return amount

    def bulk_insert(self, table, items):
        return [self.insert(table, item) for item in items]


class Expansion(object):
    """
    Class to encapsulate a pydal expression and track the parse
    expansion and its results.

    Two different MongoDB mechanisms are targeted here.  If the query
    is sufficiently simple, then simple queries are generated.  The
    bulk of the complexity here is however to support more complex
    queries that are targeted to the MongoDB Aggregation Pipeline.

    This class supports four operations: 'count', 'select', 'update'
    and 'delete'.

    Behavior varies somewhat for each operation type.  However
    building each pipeline stage is shared where the behavior is the
    same (or similar) for the different operations.

    In general an attempt is made to build the query without using the
    pipeline, and if that fails then the query is rebuilt with the
    pipeline.

    QUERY constructed in _build_pipeline_query():
      $project : used to calculate expressions if needed
      $match: filters out records

    FIELDS constructed in _expand_fields():
        FIELDS:COUNT
          $group : filter for distinct if needed
          $group: count the records remaining

        FIELDS:SELECT
          $group : implement aggregations if needed
          $project: implement expressions (etc) for select

        FIELDS:UPDATE
          $project: implement expressions (etc) for update

    HAVING constructed in _add_having():
      $project : used to calculate expressions
      $match: filters out records
      $project : used to filter out previous expression fields

    """

    def __init__(
        self,
        adapter,
        crud,
        query,
        fields=(),
        tablename=None,
        groupby=None,
        distinct=False,
        having=None,
    ):
        self.adapter = adapter
        self.NULL_QUERY = {
            "_id": {"$gt": self.adapter.ObjectId("000000000000000000000000")}
        }
        self._parse_data = {
            "pipeline": False,
            "need_group": bool(groupby or distinct or having),
        }
        self.crud = crud
        self.having = having
        self.distinct = distinct
        if not groupby and distinct:
            if distinct is True:
                # groupby gets all fields
                self.groupby = fields
            else:
                self.groupby = distinct
        else:
            self.groupby = groupby

        if crud == "update":
            self.values = [
                (f[0], self.annotate_expression(f[1])) for f in (fields or [])
            ]
            self.fields = [f[0] for f in self.values]
        else:
            self.fields = [self.annotate_expression(f) for f in (fields or [])]

        self.tablename = tablename or adapter.get_table(query, *self.fields)._tablename
        if use_common_filters(query):
            query = adapter.common_filter(query, [self.tablename])
        self.query = self.annotate_expression(query)

        # expand the query
        self.pipeline = []
        self.query_dict = adapter.expand(self.query)
        self.field_dicts = adapter.SON()
        self.field_groups = adapter.SON()
        self.field_groups["_id"] = adapter.SON()

        if self._parse_data["pipeline"]:
            # if the query needs the aggregation engine, set that up
            self._build_pipeline_query()

            # expand the fields for the aggregation engine
            self._expand_fields(None)
        else:
            # expand the fields
            try:
                if not self._parse_data["need_group"]:
                    self._expand_fields(self._fields_loop_abort)
                else:
                    self._parse_data["pipeline"] = True
                    raise StopIteration
            except StopIteration:
                # if the fields needs the aggregation engine, set that up
                self.field_dicts = adapter.SON()
                if self.query_dict:
                    if self.query_dict != self.NULL_QUERY:
                        self.pipeline = [{"$match": self.query_dict}]
                    self.query_dict = {}
                # expand the fields for the aggregation engine
                self._expand_fields(None)

        if not self._parse_data["pipeline"]:
            if crud == "update":
                # do not update id fields
                for fieldname in ("_id", "id"):
                    if fieldname in self.field_dicts:
                        del self.field_dicts[fieldname]
        else:
            if crud == "update":
                self._add_all_fields_projection(self.field_dicts)
                self.field_dicts = adapter.SON()

            elif crud == "select":
                if self._parse_data["need_group"]:
                    if not self.groupby:
                        # no groupby, aggregate all records
                        self.field_groups["_id"] = None
                    # id has no value after aggregations
                    self.field_dicts["_id"] = False
                    self.pipeline.append({"$group": self.field_groups})
                if self.field_dicts:
                    self.pipeline.append({"$project": self.field_dicts})
                    self.field_dicts = adapter.SON()
                self._add_having()

            elif crud == "count":
                if self._parse_data["need_group"]:
                    self.pipeline.append({"$group": self.field_groups})
                self.pipeline.append({"$group": {"_id": None, "count": {"$sum": 1}}})

            # elif crud == 'delete':
            #    pass

    @property
    def dialect(self):
        return self.adapter.dialect

    def _build_pipeline_query(self):
        # search for anything needing the $match stage.
        #   currently only '$regex' requires the match stage
        def parse_need_match_stage(items, parent, parent_key):
            need_match = False
            non_matched_indices = []
            if isinstance(items, list):
                indices = range(len(items))
            elif isinstance(items, dict):
                indices = items.keys()
            else:
                return

            for i in indices:
                if parse_need_match_stage(items[i], items, i):
                    need_match = True

                elif i not in [self.dialect.REGEXP_MARK1, self.dialect.REGEXP_MARK2]:
                    non_matched_indices.append(i)

                if i == self.dialect.REGEXP_MARK1:
                    need_match = True
                    self.query_dict["project"].update(items[i])
                    parent[parent_key] = items[self.dialect.REGEXP_MARK2]

            if need_match:
                for i in non_matched_indices:
                    name = str(items[i])
                    self.query_dict["project"][name] = items[i]
                    items[i] = {name: True}

            if parent is None and self.query_dict["project"]:
                self.query_dict["match"] = items
            return need_match

        expanded = self.adapter.expand(self.query)

        if self.dialect.REGEXP_MARK1 in expanded:
            # the REGEXP_MARK is at the top of the tree, so can just split
            # the regex over a '$project' and a '$match'
            self.query_dict = None
            match = expanded[self.dialect.REGEXP_MARK2]
            project = expanded[self.dialect.REGEXP_MARK1]

        else:
            self.query_dict = {"project": {}, "match": {}}
            if parse_need_match_stage(expanded, None, None):
                project = self.query_dict["project"]
                match = self.query_dict["match"]
            else:
                project = {"__query__": expanded}
                match = {"__query__": True}

        if self.crud in ["select", "update"]:
            self._add_all_fields_projection(project)
        else:
            self.pipeline.append({"$project": project})
        self.pipeline.append({"$match": match})
        self.query_dict = None

    def _expand_fields(self, mid_loop):
        if self.crud == "update":
            mid_loop = mid_loop or self._fields_loop_update_pipeline
            for field, value in self.values:
                self._expand_field(field, value, mid_loop)
        elif self.crud in ["select", "count"]:
            mid_loop = mid_loop or self._fields_loop_select_pipeline
            for field in self.fields:
                self._expand_field(field, field, mid_loop)
        elif self.fields:
            raise RuntimeError(self.crud + " not supported with fields")

    def _expand_field(self, field, value, mid_loop):
        expanded = {}
        if isinstance(field, Field):
            expanded = self.adapter.expand(value, field.type)
        elif isinstance(field, (Expression, Query)):
            expanded = self.adapter.expand(field)
            field.name = str(expanded)
        else:
            raise RuntimeError("%s not supported with fields" % type(field))

        if mid_loop:
            expanded = mid_loop(expanded, field, value)
        self.field_dicts[field.name] = expanded

    def _fields_loop_abort(self, expanded, *args):
        # if we need the aggregation engine, then start over
        if self._parse_data["pipeline"]:
            raise StopIteration()
        return expanded

    def _fields_loop_update_pipeline(self, expanded, field, value):
        if not isinstance(value, Expression):
            if self.adapter.server_version_major >= 2.6:
                expanded = {"$literal": expanded}

            # '$literal' not present in server versions < 2.6
            elif field.type in ["string", "text", "password"]:
                expanded = {"$concat": [expanded]}
            elif field.type in ["integer", "bigint", "float", "double"]:
                expanded = {"$add": [expanded]}
            elif field.type == "boolean":
                expanded = {"$and": [expanded]}
            elif field.type in ["date", "time", "datetime"]:
                expanded = {"$add": [expanded]}
            else:
                raise RuntimeError(
                    "updating with expressions not supported for field type "
                    + "'%s' in MongoDB version < 2.6" % field.type
                )
        return expanded

    def _fields_loop_select_pipeline(self, expanded, field, value):
        # search for anything needing $group
        def parse_groups(items, parent, parent_key):
            for item in items:
                if isinstance(items[item], list):
                    for list_item in items[item]:
                        if isinstance(list_item, dict):
                            parse_groups(
                                list_item, items[item], items[item].index(list_item)
                            )

                elif isinstance(items[item], dict):
                    parse_groups(items[item], items, item)

                if item == self.dialect.GROUP_MARK:
                    name = str(items)
                    self.field_groups[name] = items[item]
                    parent[parent_key] = "$" + name
            return items

        if self.dialect.AS_MARK in field.name:
            # The AS_MARK in the field name is used by base to alias the
            # result, we don't actually need the AS_MARK in the parse tree
            # so we remove it here.
            if isinstance(expanded, list):
                # AS mark is first element in list, drop it
                expanded = expanded[1]

            elif self.dialect.AS_MARK in expanded:
                # AS mark is element in dict, drop it
                del expanded[self.dialect.AS_MARK]

            else:
                # ::TODO:: should be possible to do this...
                raise SyntaxError("AS() not at top of parse tree")

        if self.dialect.GROUP_MARK in expanded:
            # the GROUP_MARK is at the top of the tree, so can just pass
            # the group result straight through the '$project' stage
            self.field_groups[field.name] = expanded[self.dialect.GROUP_MARK]
            expanded = 1

        elif self.dialect.GROUP_MARK in field.name:
            # the GROUP_MARK is not at the top of the tree, so we need to
            # pass the group results through to a '$project' stage.
            expanded = parse_groups(expanded, None, None)

        elif self._parse_data["need_group"]:
            if field in self.groupby:
                # this is a 'groupby' field
                self.field_groups["_id"][field.name] = expanded
                expanded = "$_id." + field.name
            else:
                raise SyntaxError("field '%s' not in groupby" % field)

        return expanded

    def _add_all_fields_projection(self, fields):
        for fieldname in self.adapter.db[self.tablename].fields:
            # add all fields to projection to pass them through
            if fieldname not in fields and fieldname not in ("_id", "id"):
                fields[fieldname] = 1
        self.pipeline.append({"$project": fields})

    def _add_having(self):
        if not self.having:
            return
        self._expand_field(self.having, None, self._fields_loop_select_pipeline)
        fields = {"__having__": self.field_dicts[self.having.name]}
        for fieldname in self.pipeline[-1]["$project"]:
            # add all fields to projection to pass them through
            if fieldname not in fields and fieldname not in ("_id", "id"):
                fields[fieldname] = 1

        self.pipeline.append({"$project": copy.copy(fields)})
        self.pipeline.append({"$match": {"__having__": True}})
        del fields["__having__"]
        self.pipeline.append({"$project": fields})

    def annotate_expression(self, expression):
        def mark_has_field(expression):
            if not isinstance(expression, (Expression, Query)):
                return False
            first_has_field = mark_has_field(expression.first)
            second_has_field = mark_has_field(expression.second)
            expression.has_field = (
                isinstance(expression, Field) or first_has_field or second_has_field
            )
            return expression.has_field

        def add_parse_data(child, parent):
            if isinstance(child, (Expression, Query)):
                child.parse_root = parent.parse_root
                child.parse_parent = parent
                child.parse_depth = parent.parse_depth + 1
                child._parse_data = parent._parse_data
                add_parse_data(child.first, child)
                add_parse_data(child.second, child)
            elif isinstance(child, (list, tuple)):
                for c in child:
                    add_parse_data(c, parent)

        if isinstance(expression, (Expression, Query)):
            expression.parse_root = expression
            expression.parse_depth = -1
            expression._parse_data = self._parse_data
            add_parse_data(expression, expression)
        mark_has_field(expression)
        return expression

    def get_collection(self, safe=None):
        return self.adapter._get_collection(self.tablename, safe)


class MongoBlob(Binary):
    MONGO_BLOB_BYTES = USER_DEFINED_SUBTYPE
    MONGO_BLOB_NON_UTF8_STR = USER_DEFINED_SUBTYPE + 1

    def __new__(cls, value):
        # return None and Binary() unmolested
        if value is None or isinstance(value, Binary):
            return value

        # bytearray is marked as MONGO_BLOB_BYTES
        if isinstance(value, bytearray):
            return Binary.__new__(cls, bytes(value), MongoBlob.MONGO_BLOB_BYTES)

        # return non-strings as Binary(), eg: PY3 bytes()
        if not isinstance(value, basestring):
            return Binary(value)

        # if string is encodable as UTF-8, then return as string
        try:
            value.encode("utf-8")
            return value
        except UnicodeDecodeError:
            # string which can not be UTF-8 encoded, eg: pickle strings
            return Binary.__new__(cls, value, MongoBlob.MONGO_BLOB_NON_UTF8_STR)

    def __repr__(self):
        return repr(MongoBlob.decode(self))

    @staticmethod
    def decode(value):
        if isinstance(value, Binary):
            if value.subtype == MongoBlob.MONGO_BLOB_BYTES:
                return bytearray(value)
            if value.subtype == MongoBlob.MONGO_BLOB_NON_UTF8_STR:
                return str(value)
        return value
